import re
from datetime import datetime
from typing import Self

from pystrict import strict

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd


#@strict
@serializer
class Xsd_dateTimeStamp(Xsd):
    """
    Implements the XML Schema [xsd:dateTimeStamp](https://www.w3.org/TR/xmlschema11-2/#dateTimeStamp) datatype
    """
    __value: datetime

    def __init__(self, value: datetime | Self | str, validate: bool = True):
        """
        Constructor for the Xsd_dateTimeStamp instance
        :param value: A Xsd_dateTimeStamp instance, or a datetime object or a string
        containing a datetimestmp in OS format
        :type value: datetime | Self | str
        :raises OldapErrorValue: If the value is not a datetimestamp object
        """
        if isinstance(value, Xsd_dateTimeStamp):
            self.__value = value.__value
        elif isinstance(value, datetime):
            self.__value = value
        else:
            if validate:
                if re.match(
                        r'^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])T([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\.\d+)?(Z|[+-]([01][0-9]|2[0-3]):[0-5][0-9])$',
                        value) is None:
                    raise OldapErrorValue(f'DateTimeStamp "{value}" not a valid ISO 8601.')
            try:
                self.__value = datetime.fromisoformat(value)
            except ValueError as err:
                raise OldapErrorValue(str(err))

    def __str__(self) -> str:
        """
        String representation of the Xsd_dateTimeStamp instance as ISO datetimestamp
        :return:
        """
        return self.__value.isoformat()

    def __repr__(self) -> str:
        """
        Constructor string representation of the Xsd_dateTimeStamp instance as ISO datetimestamp
        :return: Constructor string
        """
        return f'Xsd_dateTimeSTamp("{self.__value.isoformat()}")'

    def __eq__(self, other: Self | datetime | str | None) -> bool:
        """
        Equality check of the Xsd_dateTimeStamp instance with another value
        :param other: A Xsd_dateTimeStamp instance, or a datetime object or a string, or None
        :type other: Xsd_dateTimeStamp | datetime | str | None
        :return: True of False
        :raise OldapErrorValue: If the value is not a valid datetimestamp object
        """
        if other is None:
            return False
        if isinstance(other, Xsd_dateTimeStamp):
            return self.__value == other.__value
        if isinstance(other, datetime):
            return self.__value == other
        else:
            if re.match(
                    r'^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12][0-9]|3[01])T([01][0-9]|2[0-3]):([0-5][0-9]):([0-5][0-9]|60)(\.\d+)?(Z|[+-]([01][0-9]|2[0-3]):[0-5][0-9])$',
                    other) is None:
                raise OldapErrorValue(f'DateTimeStamp "{other}" not a valid ISO 8601.')
            other = datetime.fromisoformat(other)
            return self.__value == other

    def __hash__(self) -> int:
        """
        Hashing of the Xsd_dateTimeStamp instance
        :return: Hash value
        """
        return hash(self.__value)

    def _as_dict(self) -> dict:
        """
        Internal method for the JSON serialization (@serializer decorator)
        :return: dict
        """
        return {'value': self.__value.isoformat()}

    @property
    def toRdf(self) -> str:
        """
        Converts the Xsd_dateTimeStamp instance to a RDF string
        :return: RDF string
        """
        return f'"{self.__value.isoformat()}"^^xsd:dateTimeStamp'

    @property
    def value(self) -> datetime:
        """
        Converts the Xsd_dateTimeStamp instance to a datetimestamp object
        :return: datetimestamp object
        """
        return self.__value



"""
This module implements the enumerations for the permissions.
"""
from enum import Enum, unique
from typing import Self

from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.helpers.serializer import serializer


@unique
@serializer
class AdminPermission(Enum):
    """
    Administrative permissions. The following permissions are supported:

    - _AdminPermission.ADMIN_OLDAP_: Quasi root permission. This useer can do everything (**dangerous!**)
    - _AdminPermission.ADMIN_USERS_: Allows to add/modify/delete users for the project this permission is given for
    - _AdminPermission.ADMIN_RESOURCES_: Override resources permission for the resources in the given project
    - _AdminPermission.ADMIN_MODEL: Change the data model
    - _AdminPermission.ADMIN_CREATE: Create new resources in the given project
    """
    ADMIN_OLDAP = 'oldap:ADMIN_OLDAP'
    ADMIN_USERS = 'oldap:ADMIN_USERS'
    ADMIN_PERMISSION_SETS = 'oldap:ADMIN_PERMISSION_SETS'
    ADMIN_RESOURCES = 'oldap:ADMIN_RESOURCES'
    ADMIN_MODEL = 'oldap:ADMIN_MODEL'
    ADMIN_CREATE = 'oldap:ADMIN_CREATE'
    ADMIN_LISTS = 'oldap:ADMIN_LISTS'

    @property
    def toRdf(self):
        return self.value


@unique
@serializer
class DataPermission(Enum):
    """
    Data permissions are given to permission sets. These permissions determine what a user is allowed to do
    with resources. Each resource as well as each user is connected to zero, one or several permission sets.
    See [Permission concept](permission_concept).

    - _DATA_RESTRICTED_: Restricted view rights
    - _DATA_VIEW_: Full view rights
    - _DATA_EXTEND_: User may extend data, but is not allowed to modify/delete existing data
    -_DATA_UPDATE_: User may update data, but is not allowed to delete date
    -_DATA_DELETE_: User may delete data
    -_DATA_PERMISSIONS_: User may change the permission of data
    """
    DATA_RESTRICTED = 'oldap:DATA_RESTRICTED'
    DATA_VIEW = 'oldap:DATA_VIEW'
    DATA_EXTEND = 'oldap:DATA_EXTEND'
    DATA_UPDATE = 'oldap:DATA_UPDATE'
    DATA_DELETE = 'oldap:DATA_DELETE'
    DATA_PERMISSIONS = 'oldap:DATA_PERMISSIONS'

    @property
    def toRdf(self):
        return self.value

    @classmethod
    def from_string(cls, permission_string: str) -> Self:
        for member in cls:
            if f'oldap:{member.name}' == permission_string:
                return member
        raise ValueError(f'{permission_string} is not in DataPermission enum.')




import uuid
from enum import unique, Enum
from typing import Self

from oldaplib.src.helpers.oldaperror import OldapErrorValue
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName


@unique
class IriRep(Enum):
    """
    Enum to indicate form of Iri
    """
    FULL = 'full'
    QNAME = 'qname'


@serializer
class Iri(Xsd):
    """
    Implements the Iri class
    """

    __value: str
    __rep: IriRep

    def __init__(self, value: Self | Xsd_QName | Xsd_anyURI | str | None = None, validate: bool = True):
        """
        Constructor for the Iri class. If no parameter is supplied, the Iri class is initialized with
        a generated random Iri from the urn-namespace.
        :param value: an Iri value. If this parameter is omitted or None, a URN is being generated
        :type value: IriRep | Xsd_anyURI | str | None
        :raises OldapErrorValue: Invalid IRI string
        """
        if value is None:
            self.__value = uuid.uuid4().urn
            self.__rep = IriRep.FULL
            return
        if isinstance(value, Iri):
            self.__value = value.__value
            self.__rep = value.__rep
            return
        if isinstance(value, Xsd_QName):
            self.__value = str(value)
            self.__rep = IriRep.QNAME
            return
        elif isinstance(value, Xsd_anyURI):
            self.__value = str(value)
            self.__rep = IriRep.FULL
            return
        elif isinstance(value, str):
            try:
                tmp = Xsd_QName(value, validate=validate)
                self.__value = str(value)
                self.__rep = IriRep.QNAME
                return
            except:
                pass
            try:
                tmp = Xsd_anyURI(value, validate=validate)
                self.__value = str(value)
                self.__rep = IriRep.FULL
                return
            except:
                pass
            raise OldapErrorValue(f'Invalid string for IRI: "{value}"')
        else:
            raise OldapErrorValue(f'Invalid value for IRI: "{value}"')

    @classmethod
    def fromPrefixFragment(cls, prefix: Xsd_NCName | str, fragment: Xsd_NCName | str, validate: bool = True) -> Self:
        prefix = Xsd_NCName(prefix)
        fragment = Xsd_NCName(fragment)
        value = Xsd_QName(prefix, fragment, validate)
        return cls(value)

    def __str__(self) -> str:
        """
        String representation of the Iri
        :return: Iri string
        """
        return self.__value

    def __repr__(self) -> str:
        """
        String representation of the Iri as constructor string
        :return:
        """
        return f'Iri("{self.__value}")'

    def __eq__(self, other: Self | Xsd_QName | Xsd_anyURI | str) -> bool:
        if isinstance(other, Iri):
            return self.__value == other.__value
        elif isinstance(other, Xsd_QName):
            return self.__value == str(other)
        elif isinstance(other, Xsd_anyURI):
            return self.__value == str(other)
        else:
            return self.__value == str(other)

    def __hash__(self) -> int:
        return hash(self.__value)

    @property
    def toRdf(self) -> str:
        if self.__rep == IriRep.FULL:
            return f'<{self.__value}>'
        else:
            return self.__value

    def _as_dict(self) -> dict[str, str]:
        return {'value': self.__value}

    @property
    def is_qname(self) -> bool:
        """
        Checks if the Iri is an Iri QName
        :return:
        """
        return self.__rep == IriRep.QNAME

    @property
    def as_qname(self) -> Xsd_QName | None:
        if self.__rep == IriRep.QNAME:
            return Xsd_QName(self.__value, validate=False)
        else:
            return None

    @property
    def is_fulliri(self) -> bool:
        """
        Checks if the Iri is a full qualified Iri
        :return:
        """
        return self.__rep == IriRep.FULL

    @property
    def prefix(self) -> str:
        """
        Access the prefix of the QName as property
        :return: Prefix as string
        """
        if self.__rep == IriRep.FULL:
            return None
        parts = self.__value.split(':')
        return parts[0]

    @property
    def fragment(self) -> str | None:
        """
        Access the fragment as fragment of the QName as property
        :return: Fragment as string
        """
        if self.__rep == IriRep.FULL:
            return None
        parts = self.__value.split(':')
        return parts[1]


if __name__ == '__main__':
    iri = Iri("oldap:HyperHamlet\".}\nSELECT * WHERE{?s ?p ?s})#")



from dataclasses import dataclass
from datetime import datetime
from functools import partial
from typing import Dict, Self

from pystrict import strict

from oldaplib.src.enums.permissionsetattr import PermissionSetAttr
from oldaplib.src.enums.permissions import AdminPermission, DataPermission
from oldaplib.src.helpers.context import Context
from oldaplib.src.enums.action import Action
from oldaplib.src.project import Project
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorAlreadyExists, OldapErrorNoPermission, OldapError, \
    OldapErrorInconsistency, OldapErrorUpdateFailed, OldapErrorImmutable, OldapErrorNotFound
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.xsd.xsd_string import Xsd_string

PermissionSetAttrTypes = Xsd_NCName | LangString | DataPermission | None

@dataclass
class PermissionSetAttrChange:
    """
    A dataclass used to represent the changes made to a field.
    """
    old_value: PermissionSetAttrTypes
    action: Action


#@strict
class PermissionSet(Model):
    __datatypes = {
        PermissionSetAttr.ID: Xsd_NCName,
        PermissionSetAttr.LABEL: LangString,
        PermissionSetAttr.COMMENT: LangString,
        PermissionSetAttr.GIVES_PERMISSION: DataPermission,
        PermissionSetAttr.DEFINED_BY_PROJECT: Iri
    }

    __creator: Iri | None
    __created: Xsd_dateTime | None
    __contributor: Iri | None
    __modified: Xsd_dateTime | None

    __permset_iri: Iri | None

    __attributes: Dict[PermissionSetAttr, PermissionSetAttrTypes]

    __changeset: Dict[PermissionSetAttr, PermissionSetAttrChange]

    def __init__(self, *,
                 con: IConnection,
                 creator: Iri | str |None = None,
                 created: Xsd_dateTime | datetime | str | None = None,
                 contributor: Iri | None = None,
                 modified: Xsd_dateTime | datetime | str | None = None,
                 id: Xsd_NCName | str,
                 label: LangString | str | None = None,
                 comment: LangString | str | None = None,
                 givesPermission: DataPermission,
                 definedByProject: Iri | Xsd_NCName):
        """
        Constructor for a permission set.
        :param con: Subclass of IConnection
        :type con: IConnection
        :param creator: Usually not being used (internal use only)
        :type creator: Iri | None
        :param created: Usually not being used (internal use only)
        :type created: Xsd_dateTime | datetime | str | None
        :param contributor: Usually not being used (internal use only)
        :type contributor: Iri | None
        :param modified: Usually not being used (internal use only)
        :type modified: Xsd_dateTime | datetime | str | None
        :param id: A unique identifier for the permission set (unique within the project as given be :definedByProject)
        :type id: Xsd_NCName | str
        :param label: A meaninful label for the permission set (several languages allowed)
        :type label: LangString | str
        :param comment: A meaningful comment for the permission set (several languages allowed)
        :type comment: LangString | str
        :param givesPermission: The permission that this permision set grants
        :type givesPermission: DataPermission
        :param definedByProject: The project that defines this permission set (either the IRI or the shortname)
        :type definedByProject: Iri | Xsd_NCName
        :raises OldapErrorNoFound: Given project does not exist
        """
        super().__init__(con)
        self.__creator = Iri(creator) if creator else con.userIri
        self.__created = Xsd_dateTime(created) if created else None
        self.__contributor = Iri(contributor) if contributor else con.userIri
        self.__modified = Xsd_dateTime(modified) if modified else None
        self.__attributes = {}

        self.__attributes[PermissionSetAttr.ID] = Xsd_NCName(id)
        self.__attributes[PermissionSetAttr.LABEL] = LangString(label)
        self.__attributes[PermissionSetAttr.LABEL].set_notifier(self.notifier, PermissionSetAttr.LABEL)
        self.__attributes[PermissionSetAttr.COMMENT] = LangString(comment)
        self.__attributes[PermissionSetAttr.COMMENT].set_notifier(self.notifier, PermissionSetAttr.COMMENT)
        self.__attributes[PermissionSetAttr.GIVES_PERMISSION] = givesPermission
        #
        # get the project IRI
        #
        project = Project.read(self._con, definedByProject)
        self.__attributes[PermissionSetAttr.DEFINED_BY_PROJECT] = project.projectIri

        if not self.__attributes[PermissionSetAttr.ID]:
            raise OldapErrorInconsistency(f'PermissionSet must have a unique ID, none given.')
        if not self.__attributes[PermissionSetAttr.LABEL]:
            raise OldapErrorInconsistency(f'PermissionSet must have at least one rdfs:label, none given.')
        if not self.__attributes[PermissionSetAttr.GIVES_PERMISSION]:
            raise OldapErrorInconsistency(f'PermissionSet must have at least one oldap:givesPermission, none given.')
        if not self.__attributes[PermissionSetAttr.DEFINED_BY_PROJECT]:
            raise OldapErrorInconsistency(f'PermissionSet must have at least one oldap:definedByProject, none given.')

        project = Project.read(self._con, self.__attributes[PermissionSetAttr.DEFINED_BY_PROJECT])
        self.__permset_iri = Iri.fromPrefixFragment(project.projectShortName, self.__attributes[PermissionSetAttr.ID], validate=False)

        for field in PermissionSetAttr:
            prefix, name = field.value.split(':')
            setattr(PermissionSet, name, property(
                partial(self.__get_value, field=field),
                partial(self.__set_value, field=field),
                partial(self.__del_value, field=field)))
        self.__changeset = {}

    def check_for_permissions(self) -> (bool, str):
        """
        Internal method to check if a user may modify the permission set.
        :return: a tuple with a boolean (True, False) and the error message (or "OK")
        """
        #
        # First we check if the logged-in user ("actor") has the ADMIN_PERMISSION_SETS permission for
        # the given project!
        #
        actor = self._con.userdata
        sysperms = actor.inProject.get(Iri('oldap:SystemProject'))
        if sysperms and AdminPermission.ADMIN_OLDAP in sysperms:
            #
            # user has root privileges!
            #
            return True, "OK"
        else:
            if actor.inProject.get(self.definedByProject) is None:
                return False, f'Actor has no ADMIN_PERMISSION_SETS permission for project {self.definedByProject}'
            else:
                if AdminPermission.ADMIN_PERMISSION_SETS not in actor.inProject.get(self.definedByProject):
                    return False, f'Actor has no ADMIN_PERMISSION_SETS permission for project {self.definedByProject}'
            return True, "OK"

    def __get_value(self: Self, self2: Self, field: PermissionSetAttr) -> PermissionSetAttrTypes | None:
        return self.__attributes.get(field)

    def __set_value(self: Self, self2: Self, value: PermissionSetAttrTypes, field: PermissionSetAttr) -> None:
        if field == PermissionSetAttr.ID and self.__attributes.get(PermissionSetAttr.ID) is not None:
            OldapErrorAlreadyExists(f'A permission set ID already has been assigned: "{repr(self.__attributes.get(PermissionSetAttr.PERMISSION_SET_IRI))}".')
        self.__change_setter(field, value)

    def __del_value(self: Self, self2: Self, field: PermissionSetAttr) -> None:
        del self.__attributes[field]

    def __change_setter(self, attr: PermissionSetAttr, value: PermissionSetAttrTypes) -> None:
        if self.__attributes[attr] == value:
            return
        if attr in {PermissionSetAttr.ID, PermissionSetAttr.DEFINED_BY_PROJECT}:
            raise OldapErrorImmutable(f'Field {attr.value} is immutable.')
        if self.__attributes[attr] is None:
            if self.__changeset.get(attr) is None:
                self.__changeset[attr] = PermissionSetAttrChange(None, Action.CREATE)
        else:
            if value is None:
                if self.__changeset.get(attr) is None:
                    self.__changeset[attr] = PermissionSetAttrChange(self.__attributes[attr], Action.DELETE)
            else:
                if self.__changeset.get(attr) is None:
                    self.__changeset[attr] = PermissionSetAttrChange(self.__attributes[attr], Action.REPLACE)

        if value is None:
            del self.__attributes[attr]
        else:
            if isinstance(self.__datatypes[attr], set):
                dtypes = list(self.__datatypes[attr])
                for dtype in dtypes:
                    try:
                        self.__attributes[attr] = dtype(value)
                        break;
                    except OldapErrorValue:
                        pass
            else:
                self.__attributes[attr] = self.__datatypes[attr](value)

    def __str__(self) -> str:
        res = f'PermissionSet: {self.__attributes[PermissionSetAttr.ID]}\n'\
              f'  Creation: {self.__created} by {self.__creator}\n'\
              f'  Modified: {self.__modified} by {self.__contributor}\n' \
              f'  Label: {self.__attributes.get(PermissionSetAttr.LABEL, "-")}\n' \
              f'  Comment: {self.__attributes.get(PermissionSetAttr.COMMENT, "-")}\n'\
              f'  Permission: {self.__attributes[PermissionSetAttr.GIVES_PERMISSION].name}\n'\
              f'  By project: {self.__attributes[PermissionSetAttr.DEFINED_BY_PROJECT]}\n'
        return res

    def __getitem__(self, attr: PermissionSetAttr) -> PermissionSetAttrTypes:
        return self.__attributes[attr]

    def get(self, attr: PermissionSetAttr) -> PermissionSetAttrTypes:
        return self.__attributes.get(attr)

    def __setitem__(self, attr: PermissionSetAttr, value: PermissionSetAttrTypes) -> None:
        self.__change_setter(attr, value)

    def __delitem__(self, attr: PermissionSetAttr) -> None:
        if self.__attributes.get(attr) is not None:
            self.__changeset[attr] = PermissionSetAttrChange(self.__attributes[attr], Action.DELETE)
            del self.__attributes[attr]

    @property
    def creator(self) -> Xsd_anyURI | None:
        return self.__creator

    @property
    def created(self) -> datetime | None:
        return self.__created

    @property
    def contributor(self) -> Xsd_anyURI | None:
        return self.__contributor

    @property
    def modified(self) -> datetime | None:
        return self.__modified

    def notifier(self, what: PermissionSetAttr) -> None:
        self.__changeset[what] = PermissionSetAttrChange(None, Action.MODIFY)

    @property
    def changeset(self) -> Dict[PermissionSetAttr, PermissionSetAttrChange]:
        """
        Return the changeset, that is dicst with information about all properties that have benn changed.
        :return: A dictionary of all changes
        """
        return self.__changeset

    def clear_changeset(self) -> None:
        """
        Clear the changeset.
        :return: None
        """
        self.__changeset = {}

    def create(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Create the given permission set in the triple store.
        :param indent: indentation for SPARQL text
        :type indent: int
        :param indent_inc: indentation increment for the SPARQL text
        :type indent_inc: int
        :return: None
        """
        if self._con is None:
            raise OldapError("Cannot create: no connection")

        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        context = Context(name=self._con.context_name)
        blank = ''

        sparql1 = context.sparql_context
        sparql1 += f"""
        SELECT ?permset
        FROM oldap:admin
        WHERE {{
            ?permset a oldap:PermissionSet .
            FILTER(?permset = {self.__permset_iri.toRdf})       
        }}
        """

        timestamp = Xsd_dateTime()
        sparql = context.sparql_context
        sparql += f'{blank:{indent * indent_inc}}INSERT DATA {{\n'
        sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'

        sparql += f'{blank:{(indent + 2) * indent_inc}} {self.__permset_iri.toRdf} a oldap:PermissionSet'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:creator {self._con.userIri.toRdf}'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:created {timestamp.toRdf}'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:contributor {self._con.userIri.toRdf}'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}dcterms:modified {timestamp.toRdf}'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{PermissionSetAttr.LABEL.value} {self.label.toRdf}'
        if self.comment:
            sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{PermissionSetAttr.COMMENT.value} {self.comment.toRdf}'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{PermissionSetAttr.GIVES_PERMISSION.value} oldap:{self.givesPermission.name}'
        sparql += f' ;\n{blank:{(indent + 3) * indent_inc}}{PermissionSetAttr.DEFINED_BY_PROJECT.value} {self.definedByProject.toRdf}'

        sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
        sparql += f'{blank:{indent * indent_inc}}}}\n'

        self._con.transaction_start()
        try:
            jsonobj = self._con.transaction_query(sparql1)
        except OldapError:
            self._con.transaction_abort()
            raise
        res = QueryProcessor(context, jsonobj)
        if len(res) > 0:
            self._con.transaction_abort()
            raise OldapErrorAlreadyExists(f'A permission set "{self.__permset_iri}" already exists')

        try:
            self._con.transaction_update(sparql)
        except OldapError:
            self._con.transaction_abort()
            raise
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self.__created = timestamp
        self.__creator = self._con.userIri
        self.__modified = timestamp
        self.__contributor = self._con.userIri

    @classmethod
    def read(cls, con: IConnection, id: Xsd_NCName | str, definedByProject: Iri | Xsd_NCName | str) -> Self:
        """
        Reads a given permission set. The permission set is defined by its ID (which must be unique within
        one project) and the project IRI.
        :param con: A Connection object.
        :type con: IConnection
        :param id: The ID of the permission set.
        :type id: Xsd_NCName | str
        :param definedByProject: Iri or the shortname of the project
        :type definedByProject: Iri | Xsd_NCName | str
        :return: A PermissionSet instance
        :rtype: OldapPermissionSet
        :raises OldapErrorNot found: If the permission set cannot be found.
        """
        id = Xsd_NCName(id)
        definedByProject = Iri(definedByProject)

        project = Project.read(con, definedByProject)
        permset_iri = Iri.fromPrefixFragment(project.projectShortName, id, validate=False)
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        SELECT ?permset ?p ?o
        FROM oldap:admin
        WHERE {{
            BIND({permset_iri.toRdf} as ?permset)
            ?permset a oldap:PermissionSet .
            ?permset ?p ?o .
        }}
        """
        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        if len(res) == 0:
            raise OldapErrorNotFound(f'No permission set "{permset_iri}"')

        permset_iri: Iri | None = None
        creator: Iri | None = None
        created: Xsd_dateTime | None = None
        contributor: Iri | None = None
        modified: Xsd_dateTime | None = None
        label: LangString = LangString()
        comment: LangString = LangString()
        givesPermission: DataPermission | None = None
        definedByProject: Iri | None = None
        for r in res:
            if not permset_iri:
                try:
                    permset_iri = r['permset']
                except Exception as e:
                    raise OldapErrorInconsistency(f'Invalid project identifier "{r['o']}".')
            match str(r['p']):
                case 'dcterms:creator':
                    creator = r['o']
                case 'dcterms:created':
                    created = r['o']
                case 'dcterms:contributor':
                    contributor = r['o']
                case 'dcterms:modified':
                    modified = r['o']
                case 'rdfs:label':
                    label.add(r['o'])
                case 'rdfs:comment':
                    comment.add(r['o'])
                case 'oldap:givesPermission':
                    givesPermission = DataPermission.from_string(str(r['o']))
                case 'oldap:definedByProject':
                    definedByProject = r['o']
        cls.__permset_iri = permset_iri
        return cls(con=con,
                   id=id,
                   creator=creator,
                   created=created,
                   contributor=contributor,
                   modified=modified,
                   label=label,
                   comment=comment,
                   givesPermission=givesPermission,
                   definedByProject=Iri(definedByProject, validate=False))

    @staticmethod
    def search(con: IConnection, *,
               id: str | None = None,
               definedByProject: Iri | str | None = None,
               givesPermission: DataPermission | None = None,
               label: Xsd_string | str | None = None) -> list[Iri | Xsd_QName]:
        """
        Search for a permission set. At least one of the search criteria is required. Multiple search criteria are
        combined using a logical AND.
        :param con: A valid Connection object.
        :type con: IConnection
        :param id: Search for the given ID. The given string must be _contained_ in the ID (substring)
        :type id: str | None
        :param definedByProject: The project which is responsible for the permission set
        :type definedByProject: str | None
        :param givesPermission: The permission that the permission set should grant
        :type givesPermission: str | None
        :param label: The label string. The given string must be within at least one language label.
        :type label: str | None
        :return: A list or permission set Iri's (possibly as Xsd_QName
        :rtype: list[Iri | Xsd_QName]
        """
        if definedByProject:
            definedByProject = Iri(definedByProject)
        label = Xsd_string(label)
        context = Context(name=con.context_name)
        sparql = context.sparql_context
        if definedByProject:
            sparql += 'SELECT DISTINCT ?permsetIri ?namespaceIri ?projectShortName'
            context = Context(name=con.context_name)
        else:
            sparql += 'SELECT DISTINCT ?permsetIri'
        sparql += '\n'
        sparql += 'FROM oldap:admin\n'
        sparql += 'WHERE {\n'
        sparql += '   ?permsetIri rdf:type oldap:PermissionSet .\n'
        if definedByProject:
            sparql += '   ?permsetIri oldap:definedByProject ?definedByProject .\n'
            sparql += '   ?definedByProject oldap:namespaceIri ?namespaceIri .\n'
            sparql += '   ?definedByProject oldap:projectShortName ?projectShortName .\n'
        if givesPermission:
            sparql += '   ?permsetIri oldap:givesPermission ?givesPermission .\n'
        if label:
            sparql += '   ?permsetIri rdfs:label ?label .\n'
        if id or definedByProject or givesPermission or label:
            sparql += '   FILTER('
            use_and = False
            if id:
                sparql += f'CONTAINS(STR(?permsetIri), "{Xsd_string.escaping(id)}")'
                use_and = True
            if definedByProject:
                if use_and:
                    sparql += ' && '
                sparql += f'?definedByProject = {definedByProject.toRdf}'
                use_and = True
            if givesPermission:
                if use_and:
                    sparql += ' && '
                sparql += f'?givesPermission = {givesPermission.toRdf}'
                use_and = True
            if label:
                if use_and:
                    sparql += ' && '
                if label.lang:
                    sparql += f'?label = {label.toRdf}'
                else:
                    sparql += f'CONTAINS(STR(?label), "{Xsd_string.escaping(label.value)}")'
            sparql += ')\n'
        sparql += '}\n'
        jsonobj = con.query(sparql)
        res = QueryProcessor(context, jsonobj)
        permissionSets: list[Iri] = []
        for r in res:
            if definedByProject:
                #context[r['projectShortName']] = r['namespaceIri']
                psqname = r['permsetIri'].as_qname or context.iri2qname(str(r['permsetIri']), validate=False)
                permissionSets.append(psqname or r['permsetIri'])
            else:
                permissionSets.append(r['permsetIri'])
        return permissionSets

    def update(self, indent: int = 0, indent_inc: int = 4) -> None:
        """
        Update a changed permission set
        :param indent: Internal use (indent of SPARQL text)
        :type indent: int
        :param indent_inc: Internal use (indent increment of SPARQL text)
        :type indent_inc: int
        :return: None
        :rtype: None
        :raises OldapErrorUpdateFailed: Update failed
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)
        timestamp = Xsd_dateTime.now()
        context = Context(name=self._con.context_name)
        blank = ''
        sparql_list = []

        for attr, change in self.__changeset.items():
            if attr == PermissionSetAttr.LABEL or attr == PermissionSetAttr.COMMENT:
                if change.action == Action.MODIFY:
                    sparql_list.extend(self.__attributes[attr].update(graph=Xsd_QName('oldap:admin'),
                                                                      subject=self.__permset_iri,
                                                                      subjectvar='?project',
                                                                      field=Xsd_QName(attr.value)))
                if change.action == Action.DELETE or change.action == Action.REPLACE:
                    sparql = self.__attributes[attr].delete(graph=Xsd_QName('oldap:admin'),
                                                            subject=self.__permset_iri,
                                                            field=Xsd_QName(attr.value))
                    sparql_list.append(sparql)
                if change.action == Action.CREATE or change.action == Action.REPLACE:
                    sparql = self.__attributes[attr].create(graph=Xsd_QName('oldap:admin'),
                                                            subject=self.__permset_iri,
                                                            field=Xsd_QName(attr.value))
                    sparql_list.append(sparql)
                continue
            sparql = f'{blank:{indent * indent_inc}}# PermissionSet attribute "{attr.value}" with action "{change.action.value}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH oldap:admin\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?project {attr.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?project {attr.value} {self.__attributes[attr].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.__permset_iri.toRdf} as ?project)\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?project {attr.value} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        sparql = context.sparql_context
        sparql += " ;\n".join(sparql_list)

        self._con.transaction_start()
        try:
            self._con.transaction_update(sparql)
            self.set_modified_by_iri(Xsd_QName('oldap:admin'), self.__permset_iri, self.__modified, timestamp)
            modtime = self.get_modified_by_iri(Xsd_QName('oldap:admin'), self.__permset_iri)
        except OldapError:
            self._con.transaction_abort()
            raise
        if timestamp != modtime:
            self._con.transaction_abort()
            raise OldapErrorUpdateFailed("Update failed! Timestamp does not match")
        try:
            self._con.transaction_commit()
        except OldapError:
            self._con.transaction_abort()
            raise
        self.__modified = timestamp
        self.__contributor = self._con.userIri  # TODO: move creator, created etc. to Model!

    def delete(self) -> None:
        """
        Delete the given permission set.
        :return: None
        :rtype: None
        """
        result, message = self.check_for_permissions()
        if not result:
            raise OldapErrorNoPermission(message)

        context = Context(name=self._con.context_name)
        sparql = context.sparql_context
        sparql += f"""
        DELETE WHERE {{
            {self.__permset_iri.toRdf} a oldap:PermissionSet .
            {self.__permset_iri.toRdf} ?prop ?val .
        }} 
        """
        # TODO: use transaction for error handling
        self._con.update_query(sparql)




from datetime import datetime
from functools import partial
from typing import Type, Any, Self

from pystrict import strict

from oldaplib.src.connection import Connection
from oldaplib.src.datamodel import DataModel
from oldaplib.src.enums.action import Action
from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.enums.xsd_datatypes import XsdDatatypes
from oldaplib.src.helpers.langstring import LangString
from oldaplib.src.helpers.oldaperror import OldapErrorNotFound, OldapErrorValue, OldapErrorInconsistency
from oldaplib.src.iconnection import IConnection
from oldaplib.src.model import Model
from oldaplib.src.project import Project
from oldaplib.src.propertyclass import PropertyClass
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_base64binary import Xsd_base64Binary
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_byte import Xsd_byte
from oldaplib.src.xsd.xsd_date import Xsd_date
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_decimal import Xsd_decimal
from oldaplib.src.xsd.xsd_double import Xsd_double
from oldaplib.src.xsd.xsd_duration import Xsd_duration
from oldaplib.src.xsd.xsd_float import Xsd_float
from oldaplib.src.xsd.xsd_gday import Xsd_gDay
from oldaplib.src.xsd.xsd_gmonth import Xsd_gMonth
from oldaplib.src.xsd.xsd_gmonthday import Xsd_gMonthDay
from oldaplib.src.xsd.xsd_gyear import Xsd_gYear
from oldaplib.src.xsd.xsd_gyearmonth import Xsd_gYearMonth
from oldaplib.src.xsd.xsd_hexbinary import Xsd_hexBinary
from oldaplib.src.xsd.xsd_id import Xsd_ID
from oldaplib.src.xsd.xsd_idref import Xsd_IDREF
from oldaplib.src.xsd.xsd_int import Xsd_int
from oldaplib.src.xsd.xsd_language import Xsd_language
from oldaplib.src.xsd.xsd_long import Xsd_long
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_negativeinteger import Xsd_negativeInteger
from oldaplib.src.xsd.xsd_nmtoken import Xsd_NMTOKEN
from oldaplib.src.xsd.xsd_nonnegativeinteger import Xsd_nonNegativeInteger
from oldaplib.src.xsd.xsd_nonpositiveinteger import Xsd_nonPositiveInteger
from oldaplib.src.xsd.xsd_normalizedstring import Xsd_normalizedString
from oldaplib.src.xsd.xsd_positiveinteger import Xsd_positiveInteger
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_short import Xsd_short
from oldaplib.src.xsd.xsd_string import Xsd_string
from oldaplib.src.xsd.xsd_time import Xsd_time
from oldaplib.src.xsd.xsd_token import Xsd_token
from oldaplib.src.xsd.xsd_unsignedbyte import Xsd_unsignedByte
from oldaplib.src.xsd.xsd_unsignedint import Xsd_unsignedInt
from oldaplib.src.xsd.xsd_unsignedlong import Xsd_unsignedLong
from oldaplib.src.xsd.xsd_unsignedshort import Xsd_unsignedShort

ValueType = Xsd | LangString | list[Xsd]

#@strict
class ResourceInstance(Model):
    _iri: Iri
    _values: dict[str, Xsd | LangString | list[Xsd | LangString]]

    @staticmethod
    def convert2datatype(value: Any, datatype: XsdDatatypes) -> Xsd | LangString:
        match datatype:
            case XsdDatatypes.string:
                return Xsd_string(value)
            case XsdDatatypes.langString:
                return LangString(value)
            case XsdDatatypes.boolean:
                return Xsd_boolean(value)
            case XsdDatatypes.decimal:
                return Xsd_decimal(value)
            case XsdDatatypes.float:
                return Xsd_float(value)
            case XsdDatatypes.double:
                return Xsd_double(value)
            case XsdDatatypes.duration:
                return Xsd_duration(value)
            case XsdDatatypes.dateTime:
                return Xsd_date(value)
            case XsdDatatypes.dateTimeStamp:
                return Xsd_dateTime(value)
            case XsdDatatypes.time:
                return Xsd_time(value)
            case XsdDatatypes.date:
                return Xsd_date(value)
            case XsdDatatypes.gYearMonth:
                return Xsd_gYearMonth(value)
            case XsdDatatypes.gYear:
                return Xsd_gYear(value)
            case XsdDatatypes.gMonthDay:
                return Xsd_gMonthDay(value)
            case XsdDatatypes.gDay:
                return Xsd_gDay(value)
            case XsdDatatypes.gMonth:
                return Xsd_gMonth(value)
            case XsdDatatypes.hexBinary:
                return Xsd_hexBinary(value)
            case XsdDatatypes.base64Binary:
                return Xsd_base64Binary(value)
            case XsdDatatypes.anyURI:
                return Xsd_anyURI(value)
            case XsdDatatypes.QName:
                return Xsd_QName(value)
            case XsdDatatypes.normalizedString:
                return Xsd_normalizedString(value)
            case XsdDatatypes.token:
                return Xsd_token(value)
            case XsdDatatypes.language:
                return Xsd_language(value)
            case XsdDatatypes.NCName:
                return Xsd_NCName(value)
            case XsdDatatypes.NMTOKEN:
                return Xsd_NMTOKEN(value)
            case XsdDatatypes.ID:
                return Xsd_ID(value)
            case XsdDatatypes.IDREF:
                return Xsd_IDREF(value)
            case XsdDatatypes.integer:
                return Xsd_int(value)
            case XsdDatatypes.nonPositiveInteger:
                return Xsd_nonPositiveInteger(value)
            case XsdDatatypes.negativeInteger:
                return Xsd_negativeInteger(value)
            case XsdDatatypes.long:
                return Xsd_long(value)
            case XsdDatatypes.int:
                return Xsd_int(value)
            case XsdDatatypes.short:
                return Xsd_short(value)
            case XsdDatatypes.byte:
                return Xsd_byte(value)
            case XsdDatatypes.nonNegativeInteger:
                return Xsd_nonNegativeInteger(value)
            case XsdDatatypes.unsignedLong:
                return Xsd_unsignedLong(value)
            case XsdDatatypes.unsignedInt:
                return Xsd_unsignedInt(value)
            case XsdDatatypes.unsignedShort:
                return Xsd_unsignedShort(value)
            case XsdDatatypes.unsignedByte:
                return Xsd_unsignedByte(value)
            case XsdDatatypes.positiveInteger:
                return Xsd_positiveInteger(value)
            case None:
                return Iri(value)
            case _:
                raise OldapErrorValue(f'Invalid datatype "{datatype}" for value "{value}"')

    def __init__(self, *,
                 iri: Iri | None = None,
                 **kwargs):
        super().__init__(self.connection)
        self._iri = iri or Iri()
        self._values = {}
        #
        # get and transform values
        #
        for prop_iri, prop in self.properties.items():
            if kwargs.get(prop_iri.fragment):
                value = kwargs[prop_iri.fragment]
                if isinstance(value, (list, tuple, set)):  # we may have multiple values...
                    if prop.datatype == XsdDatatypes.langString:
                        self._values[prop_iri.fragment] = LangString(value)
                    else:
                        self._values[prop_iri.fragment] = [self.convert2datatype(x, prop.datatype) for x in value]
                else:
                    self._values[prop_iri.fragment] = self.convert2datatype(value, prop.datatype)

        for prop_iri in self.properties.keys():
            setattr(ResourceInstance, prop_iri.fragment, property(
                partial(ResourceInstance.__get_value, attr=prop_iri.fragment),
                partial(ResourceInstance.__set_value, attr=prop_iri.fragment),
                partial(ResourceInstance.__del_value, attr=prop_iri.fragment)))

        if not self._values.get('creator'):
            self._values['creator'] = self._con.userIri
        if not self._values.get('created'):
            self._values['created'] = Xsd_dateTime()
        if not self._values.get('contributor'):
            self._values['contributor'] = self._con.userIri
        if not self._values.get('modified'):
            self._values['modified'] = Xsd_dateTime()
        #
        # consistency and conformance tests
        #
        for prop_iri, prop in self.properties.items():
            if prop.get(PropClassAttr.MIN_COUNT):  # testing for MIN_COUNT conformance
                if prop[PropClassAttr.MIN_COUNT] > 0 and not self._values.get(prop_iri.fragment):
                    raise OldapErrorValue(f'Property {prop_iri} with MIN_COUNT={prop[PropClassAttr.MIN_COUNT]} is missing')
                elif isinstance(self._values[prop_iri.fragment], (list, tuple, set)) and len(self._values[prop_iri.fragment]) < 1:
                    raise OldapErrorValue(f'Property {prop_iri} with MIN_COUNT={prop[PropClassAttr.MIN_COUNT]} is missing')
            if prop.get(PropClassAttr.MAX_COUNT):  # testing for MAX_COUNT conformance
                if isinstance(self._values.get(prop_iri.fragment), (list, tuple, set)) and len(self._values[prop_iri.fragment]) > prop[PropClassAttr.MAX_COUNT]:
                    raise OldapErrorValue(f'Property {prop_iri} with MAX_COUNT={prop[PropClassAttr.MIN_COUNT]} has to many values (n={len(self._values[prop_iri.fragment])})')
            else:
                if self._values.get(prop_iri):
                    if isinstance(self._values.get(prop_iri), (list, tuple, set)):
                        for val in self._values.get(prop_iri.fragment):
                            self.validate_value(val, prop)
                    else:
                        self.validate_value(self._values[prop_iri.fragment], prop)

    def validate_value(self, value: ValueType, property: PropertyClass):
        if property.get(PropClassAttr.LANGUAGE_IN):  # testing for LANGUAGE_IN conformance
            if not isinstance(value, LangString):
                raise OldapErrorInconsistency(f'Property {property} with LANGUAGE_IN requires datatype rdf:langstring.')
            for lang, dummy in value.items():
                if not lang in property[PropClassAttr.LANGUAGE_IN]:
                    raise OldapErrorValue(f'Property {property} with LANGUAGE_IN={property[PropClassAttr.LANGUAGE_IN]} has invalid language "{lang.value}"')
        if property.get(PropClassAttr.UNIQUE_LANG):
            return  # TODO: LangString does not yet allow multiple entries of the same language...
        if property.get(PropClassAttr.IN):
            if not value in property[PropClassAttr.IN]:
                raise OldapErrorValue(f'Property {property} with IN={property[PropClassAttr.IN]} has invalid value "{value}"')
        if property.get(PropClassAttr.MIN_LENGTH):
            l = 0
            try:
                l = len(value)
            except TypeError:
                raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MIN_LENGTH]} has no length.')
            if l < property[PropClassAttr.MIN_LENGTH]:
                raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MIN_LENGTH]} has length "{len(value)}".')
        if property.get(PropClassAttr.MAX_LENGTH):
            l = 0
            try:
                l = len(value)
            except TypeError:
                raise OldapErrorInconsistency(f'Property {property} with MAX_LENGTH={property[PropClassAttr.MAX_LENGTH]} has no length.')
            if l > property[PropClassAttr.MAX_LENGTH]:
                raise OldapErrorInconsistency(f'Property {property} with MIN_LENGTH={property[PropClassAttr.MAX_LENGTH]} has length "{len(value)}".')
        if property.get(PropClassAttr.PATTERN):
            pass  # TODO: regex pattern match!
        if property.get(PropClassAttr.MIN_EXCLUSIVE):
            v: bool | None = None
            try:
                v = value > property[PropClassAttr.MIN_EXCLUSIVE]
            except TypeError:
                raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_EXCLUSIVE]} cannot be compared to "{value}".')
            if not v:
                raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_EXCLUSIVE]} has invalid "{value}".')
        if property.get(PropClassAttr.MIN_INCLUSIVE):
            v: bool | None = None
            try:
                v = value >= property[PropClassAttr.MIN_INCLUSIVE]
            except TypeError:
                raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_INCLUSIVE]} cannot be compared to "{value}".')
            if not v:
                raise OldapErrorInconsistency(f'Property {property} with MIN_EXCLUSIVE={property[PropClassAttr.MIN_INCLUSIVE]} has invalid "{value}".')
        if property.get(PropClassAttr.MAX_EXCLUSIVE):
            v: bool | None = None
            try:
                v = value < property[PropClassAttr.MAX_EXCLUSIVE]
            except TypeError:
                raise OldapErrorInconsistency(f'Property {property} with MAX_EXCLUSIVE={property[PropClassAttr.MAX_EXCLUSIVE]} cannot be compared to "{value}".')
            if not v:
                raise OldapErrorInconsistency(f'Property {property} with MAX_EXCLUSIVE={property[PropClassAttr.MAX_EXCLUSIVE]} has invalid "{value}".')
        if property.get(PropClassAttr.MAX_INCLUSIVE):
            v: bool | None = None
            try:
                v = value <= property[PropClassAttr.MAX_INCLUSIVE]
            except TypeError:
                raise OldapErrorInconsistency(f'Property {property} with MAX_INCLUSIVE={property[PropClassAttr.MAX_INCLUSIVE]} cannot be compared to "{value}".')
            if not v:
                raise OldapErrorInconsistency(f'Property {property} with MAX_INCLUSIVE={property[PropClassAttr.MAX_INCLUSIVE]} has invalid "{value}".')
        if property.get(PropClassAttr.LESS_THAN):
            other_value = self._values.get(property[PropClassAttr.LESS_THAN])
            if isinstance(other_value, (list, tuple, list)):
                for oval in other_value:
                    b: bool | None = None
                    try:
                        b = value < oval
                    except TypeError:
                        raise OldapErrorInconsistency(
                            f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} cannot be compared "{value} / {oval}".')
                    if not b:
                        raise OldapErrorInconsistency(
                            f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} has invalid value: "{value}" NOT LESS_THAN "{oval}".')
            else:
                b: bool | None = None
                try:
                    b = value < other_value
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} cannot be compared "{value} / {other_value}".')
                if not b:
                    raise OldapErrorInconsistency(f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} has invalid value: "{value}" NOT LESS_THAN "{other_value}".')
        if property.get(PropClassAttr.LESS_THAN_OR_EQUAL):
            other_value = self._values.get(property[PropClassAttr.LESS_THAN])
            if isinstance(other_value, (list, tuple, list)):
                for oval in other_value:
                    b: bool | None = None
                    try:
                        b = value < oval
                    except TypeError:
                        raise OldapErrorInconsistency(
                            f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} cannot be compared "{value} / {oval}".')
                    if not b:
                        raise OldapErrorInconsistency(
                            f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} has invalid value: "{value}" NOT LESS_THAN "{oval}".')
            else:
                b: bool | None = None
                try:
                    b = value < other_value
                except TypeError:
                    raise OldapErrorInconsistency(f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} cannot be compared "{value} / {other_value}".')
                if not b:
                    raise OldapErrorInconsistency(f'Property {property} with LESS_THAN={property[PropClassAttr.LESS_THAN]} has invalid value: "{value}" NOT LESS_THAN "{other_value}".')

    def __get_value(self: Self, attr: str) -> ValueType | None:
        tmp = self._values.get(attr)
        if not tmp:
            return None
        return tmp

    def __set_value(self: Self, value: ValueType, attr: str) -> None:
        self._values[attr] = value
        #self.__change_setter(attr, value)

    def __del_value(self: Self, attr: str) -> None:
        #self.__changeset[attr] = ProjectAttrChange(self.__attributes[attr], Action.DELETE)
        del self._values[attr]


#@strict
class ResourceInstanceFactory:
    _con: IConnection
    _project: Project
    _datamodel: DataModel

    def __init__(self, con: IConnection, project: Project):
        self._con = con
        self._project = project
        self._datamodel = DataModel.read(con=self._con, project=self._project)

    def createObjectInstance(self, classiri: Iri, name: str) -> Type:  ## ToDo: Get name automatically from IRI
        resclass = self._datamodel.get(classiri)
        if resclass is None:
            raise OldapErrorNotFound(f'Given Resource Class "{classiri}" not found.')
        return type(name, (ResourceInstance,), {
            'connection': self._con,
            'project': self._project,
            'classiri': classiri,
            'properties': resclass.properties})








import json
import unittest

from oldaplib.src.enums.permissions import AdminPermission
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorKey
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.in_project import InProjectClass
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_string import Xsd_string


class TestInproject(unittest.TestCase):
    def test_creation(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Iri('gaga:gugus'): set()})
        self.assertEqual(len(ip), 3)
        self.assertTrue(ip)
        self.assertFalse(InProjectClass())
        p = set()
        for proj in ip.keys():
            p.add(proj)
        self.assertEqual(p, {Iri('test:proj'), Iri('https://gaga.com/test'), Iri('gaga:gugus')})

        with self.assertRaises(OldapErrorValue) as ex:
            ip = InProjectClass({'test': {'ADMIN_MODEL'}})

        with self.assertRaises(OldapErrorValue) as ex:
            ip = InProjectClass({'test:proj': {'MODEL_T'}})

    def test_get_item(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Iri('gaga:gugus'): set()})
        self.assertEqual(ip['test:proj'], {AdminPermission.ADMIN_PERMISSION_SETS, AdminPermission.ADMIN_RESOURCES})
        self.assertEqual(ip['https://gaga.com/test'], {AdminPermission.ADMIN_MODEL})
        self.assertEqual(ip[Iri('gaga:gugus')], set())

        with self.assertRaises(OldapErrorValue) as ex:
            tmp = ip['gaga']
        with self.assertRaises(OldapErrorValue) as ex:
            tmp = ip['$<>12']
        with self.assertRaises(OldapErrorValue) as ex:
            tmp = ip[Xsd_string('test:proj')]

    def test_set_item(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Xsd_QName('gaga:gugus'): set()})
        ip['test:proj'] = {AdminPermission.ADMIN_OLDAP}
        self.assertEqual(ip['test:proj'], {AdminPermission.ADMIN_OLDAP})
        ip['https://gaga.com/test'] = {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'}
        self.assertEqual(ip['https://gaga.com/test'], {AdminPermission.ADMIN_PERMISSION_SETS, AdminPermission.ADMIN_RESOURCES})
        ip[Xsd_QName('gaga:gugus')] = {'ADMIN_MODEL'}
        self.assertEqual(ip[Xsd_QName('gaga:gugus')], {AdminPermission.ADMIN_MODEL})

        with self.assertRaises(OldapErrorValue) as ex:
            ip[Xsd_QName('gaga:gugus')] = {'GAGA'}
        with self.assertRaises(OldapErrorValue) as ex:
            ip[Xsd_QName('XYZ')] = {'ADMIN_MODEL'}

    def test_del_item(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Xsd_QName('gaga:gugus'): set()})
        del ip['test:proj']
        with self.assertRaises(OldapErrorKey) as ex:
            tmp = ip['test:proj']

        with self.assertRaises(OldapErrorKey) as ex:
            del ip[Xsd_QName('gaga:blabla')]

    def test_str(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Xsd_QName('gaga:gugus'): set()})
        projs = str(ip).strip().split('\n')
        projs = [x.strip() for x in projs]
        for projstr in projs:
            proj, permset = projstr.split(' : ')
            match proj:
                case 'test:proj':
                    self.assertEqual(permset, "['oldap:ADMIN_PERMISSION_SETS', 'oldap:ADMIN_RESOURCES']")
                case 'https://gaga.com/test':
                    self.assertEqual(permset, "['oldap:ADMIN_MODEL']")
                case 'gaga:gugus':
                    self.assertEqual(permset, "[]")
                case _:
                    raise Exception("Unexpected project")

    def test_bool(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Xsd_QName('gaga:gugus'): set()})
        self.assertTrue(bool(ip))

    def test_serialization(self):
        val = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                              'https://gaga.com/test': {'ADMIN_MODEL'},
                              Xsd_QName('gaga:gugus'): set()})
        jsonstr = json.dumps(val, default=serializer.encoder_default)
        val2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
        self.assertEqual(val, val2)

    def test_copy_eq_ne(self):
        val = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                              'https://gaga.com/test': {'ADMIN_MODEL'},
                              Xsd_QName('gaga:gugus'): set()})
        val2 = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                              'https://gaga.com/test': {'ADMIN_MODEL'},
                              Xsd_QName('gaga:gugus'): set()})
        self.assertTrue(val == val2)
        self.assertFalse(val != val2)

        val = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                              'https://gaga.com/test': {'ADMIN_MODEL'},
                              Xsd_QName('gaga:gugus'): set()})
        val2 = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                              'https://gaga.com/test': {'ADMIN_MODEL'},
                              Xsd_QName('gaga:gaga'): set()})
        self.assertFalse(val == val2)
        self.assertTrue(val != val2)

        val = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                              'https://gaga.com/test': {'ADMIN_MODEL'},
                              Xsd_QName('gaga:gugus'): {AdminPermission.ADMIN_RESOURCES}})
        val2 = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                              'https://gaga.com/test': {'ADMIN_MODEL'},
                              Xsd_QName('gaga:gugus'): {AdminPermission.ADMIN_RESOURCES, AdminPermission.ADMIN_PERMISSION_SETS}})
        self.assertFalse(val == val2)
        self.assertTrue(val != val2)


    def test_items(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Xsd_QName('gaga:gugus'): set()})
        for proj, perms in ip.items():
            self.assertTrue(proj in ['test:proj', 'https://gaga.com/test', Iri('gaga:gugus')])

    def test_keys(self):
        ip = InProjectClass({'test:proj': {AdminPermission.ADMIN_PERMISSION_SETS, 'oldap:ADMIN_RESOURCES'},
                             'https://gaga.com/test': {'ADMIN_MODEL'},
                             Iri('gaga:gugus'): set()})
        keys = set(ip.keys())
        self.assertEqual(keys, {'test:proj', 'https://gaga.com/test', Iri('gaga:gugus')})


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime, timezone
from pathlib import Path
from time import sleep

from oldaplib.src.connection import Connection
from oldaplib.src.enums.sparql_result_format import SparqlResultFormat
from oldaplib.src.helpers.context import Context
from oldaplib.src.dtypes.bnode import BNode
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorNotFound
from oldaplib.src.helpers.query_processor import QueryProcessor

def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class TestBasicConnection(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)
        cls._context = Context(name="DEFAULT")
        cls._context['test'] = NamespaceIRI("http://oldap.org/test#")
        cls._context.use('test')

        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")

        cls._connection.clear_graph(Xsd_QName('test:shacl'))
        cls._connection.clear_graph(Xsd_QName('test:onto'))
        file = project_root / 'oldaplib' / 'testdata' / 'connection_test.trig'

        cls._connection.upload_turtle(file)
        sleep(1)  # upload may take a while...

    @classmethod
    def tearDownClass(cls):
        #cls._connection.clear_graph(QName('test:shacl'))
        #cls._connection.clear_graph(QName('test:onto'))
        pass

    def test_basic_connection(self):
        con = Connection(server='http://localhost:7200',
                         repo="oldap",
                         userId="rosenth",
                         credentials="RioGrande",
                         context_name="DEFAULT")
        self.assertIsInstance(con, Connection)
        self.assertEqual(con.server, 'http://localhost:7200')
        self.assertEqual(con.repo, 'oldap')
        self.assertEqual(con.context_name, 'DEFAULT')

    def test_basic_connection_wrong_credentials(self):
        with self.assertRaises(OldapError) as ex:
            con = Connection(server='http://localhost:7200',
                             repo="oldap",
                             userId="rosenth",
                             credentials="XXX",
                             context_name="DEFAULT")
        self.assertEqual(str(ex.exception), "Wrong credentials")

    def test_basic_connection_unknown_user(self):
        with self.assertRaises(OldapErrorNotFound) as ex:
            con = Connection(server='http://localhost:7200',
                             repo="oldap",
                             userId="XXX",
                             credentials="RioGrande",
                             context_name="DEFAULT")
        self.assertEqual(str(ex.exception), "Given user not found!")

    def test_inactive_user(self):
        with self.assertRaises(OldapError) as ex:
            con = Connection(server='http://localhost:7200',
                             repo="oldap",
                             userId="bugsbunny",
                             credentials="DuffyDuck",
                             context_name="DEFAULT")
        self.assertEqual(str(ex.exception), "Wrong credentials")

    def test_basic_connection_injection_userid(self):
        with self.assertRaises(OldapError) as ex:
            con = Connection(server='http://localhost:7200',
                             repo="oldap",
                             userId="rosenth \". #\n; SELECT * {?s ?p ?o}",
                             credentials="RioGrande",
                             context_name="DEFAULT")
        self.assertEqual(str(ex.exception), 'Invalid string "rosenth ". #\n; SELECT * {?s ?p ?o}" for NCName')

    def test_basic_connection_injection_credentials(self):
        with self.assertRaises(OldapError) as ex:
            con = Connection(server='http://localhost:7200',
                             repo="oldap",
                             userId="rosenth",
                             credentials="RioGrande \". #\n; SELECT * {?s ?p ?o};",
                             context_name="DEFAULT")
        self.assertEqual(str(ex.exception), 'Wrong credentials')

    def test_token(self):
        Connection.jwtkey = "This is a very special secret, yeah!"
        con = Connection(server='http://localhost:7200',
                         repo="oldap",
                         userId="rosenth",
                         credentials="RioGrande",
                         context_name="DEFAULT")
        token = con.token
        con = Connection(server='http://localhost:7200',
                         repo="oldap",
                         token=token,
                         context_name="DEFAULT")
        self.assertEqual(con.userid, Xsd_NCName("rosenth"))
        self.assertEqual(con.userIri, Xsd_anyURI("https://orcid.org/0000-0003-1681-4036"))

        with self.assertRaises(OldapError) as ex:
            con = Connection(server='http://localhost:7200',
                             repo="oldap",
                             token=token + "X",
                             context_name="DEFAULT")
        self.assertEqual(str(ex.exception), "Wrong credentials")



    #@unittest.skip('Has to be adapted to test data...')
    def test_query(self):
        query = self._context.sparql_context
        query += """
        SELECT ?s ?o
        FROM test:shacl
        WHERE {
            ?s a sh:PropertyShape .
            ?s sh:path ?o .
        }
        """
        res = self._connection.query(query)
        expected = {
            'head': {
                'vars': ['s', 'o']
            },
            'results': {
                'bindings': [{'o': {'type': 'uri', 'value': 'http://oldap.org/test#comment'},
                              's': {'type': 'uri', 'value': 'http://oldap.org/test#commentShape'}
                              },
                             {'o': {'type': 'uri', 'value': 'http://oldap.org/test#test'},
                              's': {'type': 'uri', 'value': 'http://oldap.org/test#testShape'}
                              },
                             {'o': {'type': 'uri', 'value': 'http://oldap.org/test#enum'},
                              's': {'type': 'uri', 'value': 'http://oldap.org/test#enumShape'}
                              }]
            }
        }
        self.maxDiff = None
        self.assertDictEqual(res, expected)

    #@unittest.skip('Work in progress')
    def test_json_query(self):
        expected = {
            Iri('rdf:type'): {Iri('test:testMyRes'), Iri('sh:NodeShape')},
            Iri("rdfs:comment"): "Resource for testing...",
            Iri("rdfs:label"): {"My Resource@en", "Meine Ressource@de", "Ma Resource@fr"},
            Iri("sh:property"): Xsd_QName("test:testShape"),
            Iri("sh:closed"): Xsd_boolean(True),
            Iri("sh:targetClass"): Xsd_QName("test:testMyRes"),
            Iri("dcterms:hasVersion"): '1.0.0',
            Iri("dcterms:creator"): Iri("https://orcid.org/0000-0003-1681-4036"),
            Iri("dcterms:created"): Xsd_dateTime(datetime(2023, 11, 4, 12, 0, tzinfo=timezone.utc)),
            Iri("dcterms:contributor"):  Iri("https://orcid.org/0000-0003-1681-4036"),
            Iri("dcterms:modified"): Xsd_dateTime(datetime(2023, 11, 4, 12, 0, tzinfo=timezone.utc))
        }
        query = self._context.sparql_context
        query += """
        SELECT ?prop ?value ?oo
        FROM test:shacl
        WHERE {
            test:testMyResShape ?prop ?value .
            OPTIONAL {
                ?value rdf:rest*/rdf:first ?oo
            }
        }
        """
        res = self._connection.query(query, format=SparqlResultFormat.JSON)
        result = QueryProcessor(self._context, res)
        for r in result:
            if isinstance(r['prop'], BNode) or isinstance(r['value'], BNode):
                continue
            if r['prop'] == 'rdfs:label':
                self.assertTrue(str(r['value']) in expected[r['prop']])
            elif r['prop'] == 'rdf:type':
                self.assertTrue(r['value'] in expected[r['prop']])
            else:
                self.assertEqual(r['value'], expected[r['prop']])

    def test_update_query(self):
        query1 = self._context.sparql_context
        query1 += """
        INSERT DATA {
            GRAPH test:shacl {
                test:gaga a test:Gaga .
                test:gaga rdfs:label "GAGA"
            }
        }
        """
        self._connection.update_query(query1)
        qq1 = self._context.sparql_context
        qq1 += "SELECT ?o FROM test:shacl WHERE {test:gaga rdfs:label ?o}"
        jsonres = self._connection.query(qq1)
        res = QueryProcessor(self._context, jsonres)
        self.assertEqual(len(res), 1)
        for r in res:
            self.assertEqual(str(r['o']), "GAGA")
        query2 = self._context.sparql_context
        query2 += """
        DELETE {
            GRAPH test:shacl {
                ?s rdfs:label ?o
            }
        }
        INSERT {
            GRAPH test:shacl {
                ?s rdfs:label "GUGUS"
            }
        }
        WHERE {
            ?s a test:Gaga .
            ?s rdfs:label ?o .
        }
        """
        self._connection.update_query(query2)
        qq2 = self._context.sparql_context
        qq2 += "SELECT ?o FROM test:shacl WHERE {test:gaga rdfs:label ?o}"
        jsonres = self._connection.query(qq2)
        res = QueryProcessor(self._context, jsonres)
        self.assertEqual(len(res), 1)
        for r in res:
            self.assertEqual(str(r['o']), "GUGUS")

    def test_transaction(self):
        query = self._context.sparql_context
        query += """
        INSERT DATA {
            GRAPH test:shacl {
                test:waseliwas a test:Waseliwas .
                test:waseliwas rdfs:label "WASELIWAS"
            }
        }
        """
        self._connection.update_query(query)

        self._connection.transaction_start()
        query = self._context.sparql_context
        query += """
        SELECT ?label
        WHERE {
            GRAPH test:shacl {
                ?obj a test:Waseliwas .
                ?obj rdfs:label ?label .
            }
        }
        """
        jsonobj = self._connection.transaction_query(query)
        res = QueryProcessor(self._context, jsonobj)
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0]['label']), 'WASELIWAS')

        query2 = self._context.sparql_context
        query2 += """
        DELETE {
            GRAPH test:shacl {
                ?s rdfs:label ?o
            }
        }
        INSERT {
            GRAPH test:shacl {
                ?s rdfs:label "WASELIWAS ISCH DAS DENN AU?"
            }
        }
        WHERE {
            ?s a test:Waseliwas .
            ?s rdfs:label ?o .
        }
        """
        self._connection.transaction_update(query2)

        qq2 = self._context.sparql_context
        qq2 += "SELECT ?o FROM test:shacl WHERE {test:waseliwas rdfs:label ?o}"
        jsonobj = self._connection.transaction_query(qq2)
        res = QueryProcessor(self._context, jsonobj)
        for r in res:
            self.assertEqual(str(r['o']), "WASELIWAS ISCH DAS DENN AU?")

        self._connection.transaction_commit()

        self._connection.transaction_start()
        query3 = self._context.sparql_context
        query3 += """
        DELETE {
            GRAPH test:shacl {
                ?s rdfs:label ?o
            }
        }
        INSERT {
            GRAPH test:shacl {
                ?s rdfs:label "SHOULD NOT BE CHANGED"
            }
        }
        WHERE {
            ?s a test:Waseliwas .
            ?s rdfs:Label ?o .
        }
        """
        self._connection.transaction_update(query3)

        qq2a = self._context.sparql_context
        qq2a += "SELECT ?o FROM test:shacl WHERE {test:waseliwas rdfs:label ?o}"
        res = self._connection.transaction_query(qq2a, SparqlResultFormat.JSON)
        self.assertEqual(res['results']['bindings'][0]['o']['value'], "WASELIWAS ISCH DAS DENN AU?")

        self._connection.transaction_abort()
        qq3 = self._context.sparql_context
        qq3 += "SELECT ?o FROM test:shacl WHERE {test:waseliwas rdfs:label ?o}"
        jsonres = self._connection.query(qq2)
        res = QueryProcessor(self._context, jsonres)
        self.assertEqual(len(res), 1)
        for r in res:
            self.assertEqual(str(r['o']), "WASELIWAS ISCH DAS DENN AU?")

    def test_context(self):
        context = Context(name=self._connection.context_name)
        self.assertEqual("http://oldap.org/base#", self._context['oldap'])
        self.assertEqual("http://hyperhamlet.unibas.ch/", self._context['hyha'])


if __name__ == '__main__':
    unittest.main()



"""
This module implements the error clasees
"""
from pystrict import strict
from traceback import format_exc

#@strict
class OldapError(Exception):
    """
    No special needs besides a special class for errors in OMAS
    """
    pass


#@strict
class OldapErrorValue(OldapError):
    """
    Raised when an invalid value is passed/processed
    """
    pass


#@strict
class OldapErrorType(OldapError):
    """
    Raised when an invalid value is passed/processed
    """
    pass


#@strict
class OldapErrorKey(OldapError):
    pass


#@strict
class OldapErrorNotFound(OldapError):
    """
    Raised when something was not found
    """
    pass


#@strict
class OldapErrorAlreadyExists(OldapError):
    """
    Raised when something should be created which already exists
    """
    pass


#@strict
class OldapErrorInconsistency(OldapError):
    """
    Raised when an inconsistency is encountered
    """
    pass


#@strict
class OldapErrorUpdateFailed(OldapError):
    """
    Updating failed for some reason
    """
    pass

#@strict
class OldapErrorNoPermission(OldapError):
    pass

#@strict
class OldapErrorImmutable(OldapError):
    pass

#@strict
class OldapErrorIndex(OldapError):
    pass

#@strict
class OldapErrorInUse(OldapError):
    pass




import json
import unittest
from pathlib import Path

from oldaplib.src.connection import Connection
from oldaplib.src.dtypes.bnode import BNode
from oldaplib.src.dtypes.namespaceiri import NamespaceIRI
from oldaplib.src.dtypes.rdfset import RdfSet
from oldaplib.src.dtypes.xsdset import XsdSet
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.context import Context
from oldaplib.src.helpers.oldaperror import OldapErrorValue, OldapErrorType
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.xsd.xsd_anyuri import Xsd_anyURI
from oldaplib.src.xsd.xsd_integer import Xsd_integer
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_string import Xsd_string


def find_project_root(current_path):
    # Climb up the directory hierarchy and check for a marker file
    path = Path(current_path).absolute()
    while not (path / 'pyproject.toml').exists():
        if path.parent == path:
            # Root of the filesystem, file not found
            raise RuntimeError('Project root not found')
        path = path.parent
    return path


class MyTestCase(unittest.TestCase):

    _connection: Connection

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        project_root = find_project_root(__file__)
        cls._context = Context(name="DEFAULT")
        cls._context['test'] = NamespaceIRI("http://testing.org/datatypes#")
        cls._context.use('test')
        cls._connection = Connection(server='http://localhost:7200',
                                     repo="oldap",
                                     userId="rosenth",
                                     credentials="RioGrande",
                                     context_name="DEFAULT")
        cls._connection.clear_graph(Xsd_QName('test:test'))

    @classmethod
    def tearDownClass(cls):
        pass

    def create_triple(self, name: Xsd_NCName | str, value: Xsd):
        if not isinstance(name, Xsd_NCName):
            name = Xsd_NCName(name)
        sparql = self._context.sparql_context
        sparql += f"""
        INSERT DATA {{
            GRAPH test:test {{
                test:{name} test:prop {value.toRdf}
            }}
        }}"""
        self._connection.update_query(sparql)

    def get_triple(self, name: Xsd_NCName | str) -> Xsd:
        if not isinstance(name, Xsd_NCName):
            name = Xsd_NCName(name)
        sparql = self._context.sparql_context
        sparql += f"""
        SELECT ?value
        FROM test:test
        WHERE {{
            test:{name} test:prop ?value
        }}
        """
        result = self._connection.query(sparql)
        res = QueryProcessor(context=self._context, query_result=result)
        return res[0]['value']

    def delete_triple(self, name: Xsd_NCName):
        sparql = self._context.sparql_context
        sparql += f"""
        DELETE FROM test:test {{
            test:{name} test:prop ?value
        }}
        """

    def test_namespace(self):
        ns1 = NamespaceIRI('http://www.org/test/')
        self.assertEqual(str(ns1), 'http://www.org/test/')
        self.assertEqual(repr(ns1), 'NamespaceIRI("http://www.org/test/")')
        self.assertEqual(ns1.toRdf, '"http://www.org/test/"^^xsd:anyURI')
        self.assertEqual(ns1 + "gaga", Xsd_anyURI("http://www.org/test/gaga"))

        ns2 = NamespaceIRI('http://www.org/test#')
        self.assertEqual(str(ns2), 'http://www.org/test#')
        self.assertEqual(ns2 + "gaga", Xsd_anyURI("http://www.org/test#gaga"))

        jsonstr = json.dumps(ns1, default=serializer.encoder_default)
        val2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
        self.assertEqual(ns1, val2)

        self.create_triple("NamespaceIRI", ns1)
        valx = self.get_triple("NamespaceIRI")
        self.assertIsInstance(valx, Xsd_anyURI)
        if isinstance(valx, Xsd_anyURI):
            self.assertEqual(ns1, NamespaceIRI(valx))

        with self.assertRaises(OldapErrorValue) as ex:
            nons = NamespaceIRI('http://www.org/test')
        self.assertEqual(str(ex.exception), "NamespaceIRI must end with '/' or '#'!")

        with self.assertRaises(OldapErrorValue) as ex:
            nons = NamespaceIRI('http://www.org/test\"; SELECT * FROM {?s ?p ?o}')

        with self.assertRaises(OldapErrorValue) as ex:
            nons = NamespaceIRI('http://www.org/test<super>\"; SELECT * FROM {?s ?p ?o}')

    def test_xsd_set(self):
        val = XsdSet({Xsd_string("was"), Xsd_string("ist"), Xsd_string("das?")})
        self.assertTrue(Xsd_string("was") in val)
        self.assertTrue(Xsd_string("ist") in val)
        self.assertTrue(Xsd_string("das?") in val)

        jsonstr = json.dumps(val, default=serializer.encoder_default)
        val2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
        self.assertEqual(val, val2)

        val.add(Xsd_integer(42))
        self.assertTrue(Xsd_integer(42) in val)

        val.discard(Xsd_string("ist"))
        self.assertFalse(Xsd_string("ist") in val)

        s = str(val)
        s = s[1:-1]
        s = set(s.split(", "))
        self.assertEqual(s, {"was", "42", "das?"})

        s = repr(val)
        s = s.removeprefix('XsdSet')
        s = s[1:-1]
        s = set(s.split(", "))
        self.assertEqual(s, {'Xsd_string("was")', 'Xsd_integer(42)', 'Xsd_string("das?")'})

        s = val.toRdf
        s = s[1:-1]
        s = set(s.split(" "))

        self.assertEqual(s, {'"was"^^xsd:string', '"42"^^xsd:integer', '"das?"^^xsd:string'})

        val = XsdSet(Xsd_string("was"), Xsd_string("ist"), Xsd_string("das?"))
        self.assertTrue(Xsd_string("was") in val)
        self.assertTrue(Xsd_string("ist") in val)
        self.assertTrue(Xsd_string("das?") in val)

        val = XsdSet(value={Xsd_string("was"), Xsd_string("ist"), Xsd_string("das?")})
        self.assertTrue(Xsd_string("was") in val)
        self.assertTrue(Xsd_string("ist") in val)
        self.assertTrue(Xsd_string("das?") in val)


        with self.assertRaises(OldapErrorType) as ex:
            val = XsdSet(3.5)

        with self.assertRaises(OldapErrorType) as ex:
            val = XsdSet(value=3.5)

        with self.assertRaises(OldapErrorType) as ex:
            val = XsdSet({3.5, 25})

        with self.assertRaises(OldapErrorType) as ex:
            val = XsdSet(value={3.5, 25})

        with self.assertRaises(OldapErrorType) as ex:
            val = XsdSet(5, 3.7)

    def test_bnode(self):
        val = BNode("_:node42")
        self.assertEqual(str(val), "_:node42")

        jsonstr = json.dumps(val, default=serializer.encoder_default)
        val2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
        self.assertEqual(val, val2)

    def test_string_literal(self):
        val = Xsd_string("This is a test")
        self.assertEqual(val.value, "This is a test")
        self.assertIsNone(val.lang)

        jsonstr = json.dumps(val, default=serializer.encoder_default)
        val2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
        self.assertEqual(val, val2)

        self.create_triple(Xsd_NCName("LiteralStringA"), val)
        valx = self.get_triple(Xsd_NCName("LiteralStringA"))
        self.assertEqual(val, valx)


        val = Xsd_string("This is a test", Language.EN)
        self.assertEqual(val.value, "This is a test")
        self.assertEqual(val.lang, Language.EN)

        jsonstr = json.dumps(val, default=serializer.encoder_default)
        val2 = json.loads(jsonstr, object_hook=serializer.decoder_hook)
        self.assertEqual(val, val2)

        self.create_triple(Xsd_NCName("LiteralStringB"), val)
        valx = self.get_triple(Xsd_NCName("LiteralStringB"))
        self.assertEqual(val, valx)

        val = Xsd_string("This is a test", "en")
        self.assertEqual(val.value, "This is a test")
        self.assertEqual(val.lang, Language.EN)

        with self.assertRaises(OldapErrorValue) as err:
            val = Xsd_string("This is a test", "gaga")



if __name__ == '__main__':
    unittest.main()



from enum import unique, Enum


@unique
class ResClassAttribute(Enum):
    SUPERCLASS = 'oldap:superclass'  # virtual attribute, SHACL: sh:node, OWL: rdfs:subClassOf
    LABEL = 'rdfs:label'
    COMMENT = 'rdfs:comment'
    CLOSED = 'sh:closed'



import unittest
from datetime import datetime

from oldaplib.src.enums.propertyclassattr import PropClassAttr
from oldaplib.src.enums.action import Action
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.helpers.langstring import LangString, LangStringChange
from oldaplib.src.enums.language import Language
from oldaplib.src.helpers.oldaperror import OldapError, OldapErrorValue


class TestLangstring(unittest.TestCase):

    def test_langstring_constructor(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString("english@en")
        self.assertEqual(ls1['en'], 'english')
        ls2 = LangString('nolanguage')
        self.assertEqual(ls2['zu'], 'nolanguage')
        self.assertEqual(ls2[Language.ZU], 'nolanguage')
        ls3 = LangString(ls1)
        self.assertEqual(ls3['en'], 'english')

        ls4 = LangString({
            'en': 'english',
            'de': 'deutsch',
            'fr': 'français'
        })
        self.assertEqual(ls4['fr'], 'français')
        self.assertEqual(ls4[Language.FR], 'français')
        self.assertEqual(ls4['de'], 'deutsch')
        self.assertEqual(ls4[Language.DE], 'deutsch')
        self.assertEqual(ls4['en'], 'english')
        with self.assertRaises(OldapError) as ex:
            impossible = ls4['rr']
        self.assertEqual(str(ex.exception), 'Language "rr" is invalid')

        ls5 = LangString(['english@en', 'deutsch@de', 'français@fr', 'no language'])
        self.assertEqual(ls5['fr'], 'français')
        self.assertEqual(ls5[Language.FR], 'français')
        self.assertEqual(ls5['de'], 'deutsch')
        self.assertEqual(ls5[Language.DE], 'deutsch')
        self.assertEqual(ls5['en'], 'english')
        self.assertEqual(ls5['yi'], 'english')
        with self.assertRaises(OldapError) as ex:
            impossible = ls5['rr']
        self.assertEqual(str(ex.exception), 'Language "rr" is invalid')

        ls6 = LangString({
            Language.EN: 'english',
            Language.DE: 'deutsch',
            Language.FR: 'français',
        })
        self.assertEqual(ls6['fr'], 'français')
        self.assertEqual(ls6[Language.FR], 'français')
        self.assertEqual(ls6['de'], 'deutsch')
        self.assertEqual(ls6[Language.DE], 'deutsch')
        self.assertEqual(ls6['en'], 'english')
        self.assertEqual(ls6['yi'], 'english')
        with self.assertRaises(OldapError) as ex:
            impossible = ls6['rr']
        self.assertEqual(str(ex.exception), 'Language "rr" is invalid')

        ls7 = LangString("xyz@ur")
        self.assertEqual(ls7[Language.YI], '--no string--')

        ls8 = LangString("lukas.rosenthaler@unibas.ch@en")
        ls8[Language.DE] = 'lukas.rosenthaler@gmail.com'
        self.assertEqual(ls8['en'], 'lukas.rosenthaler@unibas.ch')
        self.assertEqual(ls8['de'], 'lukas.rosenthaler@gmail.com')

        with self.assertRaises(OldapErrorValue) as ex:
            ls9 = LangString(255)

        ls10 = LangString("lukas.rosenthaler@gmail.com")
        self.assertEqual(str(ls10), '"lukas.rosenthaler@gmail.com@zu"')

    def test_langstring_empty(self):
        ls1 = LangString()
        self.assertFalse(ls1)
        self.assertEqual(ls1[Language.EN], '--no string--')

        ls2 = LangString("wasistdas@en", "", "soso@fr")
        self.assertEqual(len(ls2), 2)

        ls3 = LangString("soso@fr", "gaga@fr", "")
        self.assertEqual(len(ls3), 1)
        self.assertEqual(ls3["fr"], "gaga")

        ls4 = LangString()
        self.assertFalse(ls4)

    def test_langstring_setitem(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de", "unbekannt"])
        ls1['fr'] = 'français'
        self.assertEqual(ls1[Language.FR], 'français')

        ls2 = LangString("english@en", "deutsch@de")
        ls2[Language.FR] = 'français'
        self.assertEqual(ls2[Language.FR], 'français')

        ls3 = LangString("english@en", "deutsch@de")
        with self.assertRaises(OldapError) as ex:
            ls3['rr'] = 'no way'

    def test_langstring_undo(self):
        LangString.setDefaultLang(Language.ZU)
        ls1 = LangString(["english@en", "deutsch@de", "unbekannt"])
        ls1['fr'] = 'français'
        del ls1[Language.DE]
        self.assertIsNone(ls1.get(Language.DE))
        self.assertEqual(ls1[Language.FR], 'français')
        self.assertEqual(ls1[Language.EN], 'english')
        self.assertEqual(ls1[Language.ZU], 'unbekannt')
        ls1.undo()
        self.assertIsNone(ls1.get(Language.FR))
        self.assertEqual(ls1[Language.EN], 'english')
        self.assertEqual(ls1[Language.DE], 'deutsch')
        self.assertEqual(ls1[Language.ZU], 'unbekannt')

    def test_langstring_delete(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de", "unbekannt"])
        del ls1['de']
        self.assertEqual(ls1.changeset, {Language.DE: LangStringChange("deutsch", Action.DELETE)})
        self.assertEqual(ls1['en'], 'english')
        self.assertEqual(ls1['zu'], 'unbekannt')
        self.assertEqual(ls1['de'], 'english')

        ls2 = LangString(["english@en", "deutsch@de", "unbekannt"])
        with self.assertRaises(OldapError) as ex:
            del ls2['it']
        self.assertEqual(str(ex.exception), 'No language string of language: "it"!')

        ls3 = LangString(["english@en", "deutsch@de", "unbekannt"])
        with self.assertRaises(OldapError) as ex:
            del ls3['rr']
        self.assertEqual(str(ex.exception), 'No language string of language: "rr"!')

    def test_langstring_str(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de", "unbekannt"])
        s = str(ls1)
        s = set(s.split(", "))
        self.assertTrue('"english@en"' in s)
        self.assertTrue('"deutsch@de"' in s)
        self.assertTrue('"unbekannt@zu"' in s)

    def test_langstring_eq_ne(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de", "unbekannt"])
        ls2 = LangString(["english@en", "deutsch@de", "unbekannt"])
        self.assertTrue(ls1 == ls2)
        self.assertFalse(ls1 != ls2)
        ls3 = LangString(["english@en", "français", "unbekannt"])
        self.assertFalse(ls1 == ls3)
        self.assertTrue(ls1 != ls3)
        ls4 = LangString(["english@en", "unbekannt"])
        self.assertFalse(ls1 == ls4)
        self.assertTrue(ls1 != ls4)
        ls5 = LangString(["english@en", "deutsch@de", "français@fr", "unbekannt"])
        self.assertFalse(ls1 == ls5)
        self.assertTrue(ls1 != ls5)

    def test_langstring_items(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de", "unbekannt"])
        expected = {
            'English': 'english',
            'German': 'deutsch',
            'Zulu': 'unbekannt'
        }
        res = {}
        for lang, value in ls1.items():
            res[lang.value] = value
        self.assertDictEqual(res, expected)

    def test_langstring_langstring(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de", "unbekannt"])
        expected = {
            Language.EN: 'english',
            Language.DE: 'deutsch',
            Language.ZU: 'unbekannt'
        }
        self.assertDictEqual(ls1.langstring, expected)

    def test_langstring_add(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de"])
        ls1.add("français@fr")
        self.assertEqual(ls1, LangString("english@en", "deutsch@de", "français@fr"))
        self.assertEqual(ls1.changeset, {Language.FR: LangStringChange(None, Action.CREATE)})

        ls2 = LangString(["english@en", "deutsch@de"])
        ls2.add(["français@fr", "undefined"])
        self.assertEqual(ls2, LangString("english@en", "deutsch@de", "français@fr", "undefined@zu"))
        self.assertEqual(ls2.changeset, {Language.FR: LangStringChange(None, Action.CREATE),
                                         Language.ZU: LangStringChange(None, Action.CREATE)})

        ls3 = LangString(["english@en", "deutsch@de"])
        ls3.add({Language.FR: "français", Language.ZU: "undefined"})
        self.assertEqual(ls3, LangString("english@en", "deutsch@de", "français@fr", "undefined@zu"))
        self.assertEqual(ls2.changeset, {Language.FR: LangStringChange(None, Action.CREATE),
                                         Language.ZU: LangStringChange(None, Action.CREATE)})

        ls4 = LangString(["english@en", "deutsch@de", "französisch@fr"])
        ls4.add({Language.FR: "français", Language.ZU: "undefined"})
        self.assertEqual(ls4, LangString("english@en", "deutsch@de", "français@fr", "undefined@zu"))
        self.assertEqual(ls4.changeset, {Language.FR: LangStringChange("französisch", Action.REPLACE),
                                         Language.ZU: LangStringChange(None, Action.CREATE)})

    def test_langstring_undo(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de"])
        ls1.add({Language.FR: "français", Language.ZU: "undefined"})
        self.assertEqual(ls1, LangString("english@en", "deutsch@de", "français@fr", "undefined@zu"))
        ls1.undo()
        self.assertEqual(ls1, LangString("english@en", "deutsch@de"))

    def test_langstring_update(self):
        LangString.defaultLanguage = Language.ZU
        ls1 = LangString(["english@en", "deutsch@de"])
        ls1.add({Language.FR: "français", Language.ZU: "undefined"})
        del ls1[Language.EN]
        qlist = ls1.update(graph=Xsd_QName("oldaplib:test"),
                           subject=Iri("oldaplib:subj"),
                           subjectvar="?subj",
                           field=Xsd_QName("oldaplib:prop"))
        qstr = " ;\n".join(qlist)
        expected = """INSERT DATA {
    GRAPH oldaplib:test {
        oldaplib:subj oldaplib:prop "français"@fr .
    }
}
 ;
INSERT DATA {
    GRAPH oldaplib:test {
        oldaplib:subj oldaplib:prop "undefined"@zu .
    }
}
 ;
DELETE DATA {
    GRAPH oldaplib:test {
        oldaplib:subj oldaplib:prop "english"@en .
    }
}
"""
        self.assertEqual(qstr, expected)

        sstr = ls1.update_shacl(graph=Xsd_NCName("test"),
                                prop_iri=Iri('oldaplib:prop'),
                                attr=PropClassAttr.NAME,
                                modified=Xsd_dateTime("2023-11-04T12:00:00Z"))
        expected = """# LangString: Process "FR" with Action "create"
WITH test:shacl
INSERT {
    ?prop sh:name "français"@fr .
}
WHERE {
    BIND(oldaplib:propShape as ?prop) .
    ?prop dcterms:modified ?modified .
    FILTER(?modified = "2023-11-04T12:00:00+00:00"^^xsd:dateTime)
};
# LangString: Process "ZU" with Action "create"
WITH test:shacl
INSERT {
    ?prop sh:name "undefined"@zu .
}
WHERE {
    BIND(oldaplib:propShape as ?prop) .
    ?prop dcterms:modified ?modified .
    FILTER(?modified = "2023-11-04T12:00:00+00:00"^^xsd:dateTime)
};
# LangString: Process "EN" with Action "delete"
WITH test:shacl
DELETE {
    ?prop sh:name "english"@en .
}
WHERE {
    BIND(oldaplib:propShape as ?prop) .
    ?prop sh:name "english"@en .
    ?prop dcterms:modified ?modified .
    FILTER(?modified = "2023-11-04T12:00:00+00:00"^^xsd:dateTime)
}"""
        self.assertEqual(sstr, expected)







"""
# UserDataclass

This module implements the UserDataclass class which can be used to represent all
information about a user in the oldap universe.

The class internally represents the properties of the given user in a hidden dict.
The properties however can be accessed directly as if the class itself is a dict
or using "virtual" properties which are implemented as dynamically created getter,
setter and deleter methods.

## Use of UserDataclass

1. The UserDataclass is the base class for the `User`-class which is usually the standard
   way to access a OLDAP user. The `User`-class, in contrary to the `UserDataclass`, implements
   the CRUD operations.
2. The UserDataclass is used by the `Connection`-class for checking the authorization of the user.
   During the construction of the `Connection`-instance it must rely on direct access to the triple
   store to retrieve the user's data. The instance of the `UserDataclass` is then serialized in a
   JSON Web token.

## Other classes used by the UserDataclass

- [InProjectClass](/python_docstrings/datatypes#oldap.src.in_project):
  Defines the user's membership to projects and the administrative permission within these projects
- [RdfSet](/python_docstrings/datatypes#oldap.src.dtypes/rdfset):)


## Helper classes

- `UserFieldChange`: Bookkeeping of changes to the fields
- `UserFields`: Enum class of the data fields provided by the `UserDataclass`

"""
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from typing import Dict, Self, Set, Tuple, Any

import bcrypt

from oldaplib.src.enums.userdataclassattr import UserAttr
from oldaplib.src.helpers.context import Context
from oldaplib.src.enums.action import Action
from oldaplib.src.xsd.iri import Iri
from oldaplib.src.xsd.xsd_boolean import Xsd_boolean
from oldaplib.src.xsd.xsd_qname import Xsd_QName
from oldaplib.src.xsd.xsd_ncname import Xsd_NCName
from oldaplib.src.xsd.xsd_datetime import Xsd_dateTime
from oldaplib.src.xsd.xsd_string import Xsd_string
from oldaplib.src.xsd.xsd import Xsd
from oldaplib.src.helpers.observable_set import ObservableSet
from oldaplib.src.helpers.oldaperror import OldapErrorAlreadyExists, OldapErrorValue, OldapErrorNotFound
from oldaplib.src.enums.permissions import AdminPermission
from oldaplib.src.helpers.query_processor import QueryProcessor
from oldaplib.src.helpers.serializer import serializer
from oldaplib.src.in_project import InProjectClass


UserAttrTypes = Xsd | ObservableSet[Iri] | InProjectClass | str | bool | None


@dataclass
class UserAttrChange:
    """
    A dataclass used to represent the changes made to a field.
    """
    old_value: UserAttrTypes
    action: Action


@serializer
class UserDataclass:
    """
    The UserDataclass class implements the main handling of the user data of an oldap user. The
    UserDataclass class is the base class for the User class which implements the CRUD operations
    with the RDF store. Therefore, the UserDataclass class does not inherit from model since it
    does not need the connection instance. The class internally represents the properties of the
    given user in a hidden dict. The properties however can be accessed directly as if the class
    itself is a dict or using "virtual" properties which are implemented as dynamically created
    getter, setter and deleter methods.

    The User class inherits the following properties from the UserDataclass class:

    - `userIri: Iri` [mandatory]: IRI of the user, cannot be changed (RDF property `oldap:userIri`).
      If available, the [ORCID](https://orcid.org) should be used as Iri.
    - `userId_: Xsd_NCname | str` [mandatory]: User ID as NCName (RDF property `oldap:). Must be unique!
    - `familyName: Xsd_string | str` [mandatory]: Family name as str (RDF property `foaf:familyName`)
    - `givenName: Xsd_string | str` [mandatory]: Given name or first name as str(RDF property `foaf:givenName`)
    - `credentials: Xsd_string | str` [mandatory]: Credential (password) (RDF property `oldap:credentials`)
    - `active: Xsd_boolean | bool | None` [optional]: Is the user isActive as bool? (RDF property `oldap:isActive`)
    - `inProject: Dict[Iri | str, Set[AdminPermission]] | None `: Membership to projects and administrative permissions for this project (RDF property `oldap:inProject)
    - _hsPermission_: Permissions for data as sets of QNames (RDF property `oldap:hasPermissions`)

    These properties can be accessed as normal python class properties or using the dictionary syntax. The keys
    are defined in the [UserFields](/python_docstrings/userdataclass/#oldap.src.user_dataclass.UserFields) Enum class.
    Example for access as property:
    ```python
    user.familyName = 'Rosenthaler'
    givenname = user.givenName
    del user.givenName
    ```
    Example for access as dictionary:
    ```python
    user[UserFields.FAMILY_NAME] = 'Rosenthaler'
    givenname = user[UserFields.GIVEN_NAME]
    del user[UserFields.GIVEN_NAME]
    ```
    The class implements the following methods:

    - *UserDataclass(...)*: Constructor method for the class
    - *str()*: String representation of the class for printing
    - *add_project_permission(...)*: If the project does already exists, adds the given permission.
      If the project does not exist, it also adds the project.
    - *remove_project_permission(...)*: Removes the given permission
    """
    __datatypes = {
        UserAttr.USER_IRI: Iri,
        UserAttr.USER_ID: Xsd_NCName,
        UserAttr.FAMILY_NAME: Xsd_string,
        UserAttr.GIVEN_NAME: Xsd_string,
        UserAttr.CREDENTIALS: Xsd_string,
        UserAttr.ACTIVE: Xsd_boolean,
        UserAttr.IN_PROJECT: dict,
        UserAttr.HAS_PERMISSIONS: ObservableSet[Iri]
    }

    _creator: Iri | None
    _created: Xsd_dateTime | None
    _contributor: Iri | None
    _modified: Xsd_dateTime | None

    __attr: Dict[UserAttr, UserAttrTypes]

    __changeset: Dict[UserAttr, UserAttrChange]

    def __init__(self, *,
                 creator: Iri | str | None = None,
                 created: Xsd_dateTime | datetime | str | None = None,
                 contributor: Iri | str | None = None,
                 modified: Xsd_dateTime | datetime | str | None = None,
                 userIri: Iri | str | None = None,
                 userId: Xsd_NCName | str | None = None,
                 familyName: Xsd_string | str | None = None,
                 givenName: Xsd_string | str | None = None,
                 credentials: Xsd_string | str | None = None,
                 isActive: Xsd_boolean | bool | None = None,
                 inProject: Dict[Iri | str, Set[AdminPermission]] | None = None,
                 hasPermissions: Set[Iri] | None = None):
        """
        Constructs a new UserDataclass
        :param creator: AnyIRI of the creator of this UserDataclass
        :param created: datetime of the creation of this UserDataclass
        :param contributor: AnyIRI of the contributor of this UserDataclass
        :param modified: datetime of the modification of this UserDataclass
        :param userIri: AnyIRI of the User to be used for this UserDataclass
        :param userId: A unique identifier for the user (NCName, must be unique)
        :param familyName: The foaf:family name of the User
        :param givenName: The foaf:givenname of the User
        :param credentials: The password
        :param isActive: A boolean indicating if the user is active
        :param inProject: an InProjectType instance for the project permissions of the user
        :param hasPermissions: Set of Administrative Permissions for the user
        """
        self.__attr = {}
        self.__changeset = {}
        if inProject:
            inProjectTmp = InProjectClass(inProject, self.__inProject_cb)
        else:
            inProjectTmp = InProjectClass()
        if not isinstance(hasPermissions, ObservableSet):
            hasPermissions = ObservableSet(hasPermissions, on_change=self.__hasPermission_cb)
        if credentials is not None:
            salt = bcrypt.gensalt()
            credentials = Xsd_string(bcrypt.hashpw(str(credentials).encode('utf-8'), salt).decode('utf-8'))

        self._creator = Iri(creator) if creator else None
        self._created = Xsd_dateTime(created) if created is not None else None
        self._contributor = Iri(contributor) if contributor else None
        self._modified = Xsd_dateTime(modified) if modified else None
        self.__attr[UserAttr.USER_IRI] = Iri(userIri) if userIri else None
        self.__attr[UserAttr.USER_ID] = Xsd_NCName(userId) if userId else None
        self.__attr[UserAttr.FAMILY_NAME] = Xsd_string(familyName) if familyName else None
        self.__attr[UserAttr.GIVEN_NAME] = Xsd_string(givenName) if givenName else None
        self.__attr[UserAttr.CREDENTIALS] = Xsd_string(credentials) if credentials else None
        self.__attr[UserAttr.ACTIVE] = Xsd_boolean(isActive) if isActive is not None else None
        self.__attr[UserAttr.IN_PROJECT] = inProjectTmp if inProjectTmp is not None else None
        self.__attr[UserAttr.HAS_PERMISSIONS] = hasPermissions if hasPermissions is not None else None
        #
        # here we dynamically generate class properties for the UserFields.
        # This we can access these properties either a Dict or as property
        # for get, set and sel:
        # - user[UserFields.USER_ID]
        # - user.userId
        #
        for field in UserAttr:
            prefix, name = field.value.split(':')
            setattr(UserDataclass, name, property(
                partial(UserDataclass.__get_value, field=field),
                partial(UserDataclass.__set_value, field=field),
                partial(UserDataclass.__del_value, field=field)))
        self.clear_changeset()
    #
    # these are the methods for the getter, setter and deleter
    #
    def __get_value(self: Self, field: UserAttr) -> UserAttrTypes | None:
        return self.__attr.get(field)

    def __set_value(self: Self, value: UserAttrTypes, field: UserAttr) -> None:
        if field == UserAttr.CREDENTIALS:
            salt = bcrypt.gensalt()
            value = bcrypt.hashpw(str(value).encode('utf-8'), salt).decode('utf-8')
        if field == UserAttr.USER_IRI and self.__attr.get(UserAttr.USER_IRI) is not None:
            OldapErrorAlreadyExists(f'A user IRI already has been assigned: "{self.__attr.get(UserAttr.USER_IRI)}".')
        self.__change_setter(field, value)

    def __del_value(self: Self, field: UserAttr) -> None:
        if self.__changeset.get(field) is None:
            self.__changeset[field] = UserAttrChange(self.__attr[field], Action.DELETE)
        del self.__attr[field]
        if field == UserAttr.IN_PROJECT:
            self.__attr[field] = InProjectClass(on_change=self.__inProject_cb)
        elif field == UserAttr.HAS_PERMISSIONS:
            self.__attr[field] = ObservableSet(on_change=self.__hasPermission_cb)

    def __str__(self) -> str:
        """
        Create a string representation for the human reader
        :return: Multiline string
        """
        admin_permissions = {}
        for proj, permissions in self.__attr[UserAttr.IN_PROJECT].items():
            admin_permissions[str(proj)] = [str(x.value) for x in permissions]
        return \
            f'Userdata for {self.__attr[UserAttr.USER_IRI]}:\n' \
            f'  Creator: {self._creator}\n' \
            f'  Created at: {self._created}\n' \
            f'  Modified by: {self._contributor}\n' \
            f'  Modified at: {self._modified}\n' \
            f'  User id: {self.__attr[UserAttr.USER_ID]}\n' \
            f'  Family name: {str(self.__attr[UserAttr.FAMILY_NAME])}\n' \
            f'  Given name: {str(self.__attr[UserAttr.GIVEN_NAME])}\n' \
            f'  Active: {self.__attr[UserAttr.ACTIVE]}\n' \
            f'  In project: {admin_permissions}\n' \
            f'  Has permissions: {self.__attr[UserAttr.HAS_PERMISSIONS]}\n'

    #
    # The fields of the class can either be accessed using the dict-semantic or as
    # named properties. Here we implement the dict semantic
    #
    def __getitem__(self, item: UserAttr) -> UserAttrTypes:
        if isinstance(self.__attr.get(item), Xsd_string):
            return str(self.__attr[item])
        return self.__attr.get(item)

    def __setitem__(self, field: UserAttr, value: UserAttrTypes) -> None:
        if field == UserAttr.CREDENTIALS:
            salt = bcrypt.gensalt()
            value = bcrypt.hashpw(str(value).encode('utf-8'), salt).decode('utf-8')
        if field == UserAttr.USER_IRI and self.__attr.get(UserAttr.USER_IRI) is not None:
            OldapErrorAlreadyExists(f'A user IRI already has been assigned: "{self.__attr.get(UserAttr.USER_IRI)}".')
        self.__change_setter(field, value)

    def _as_dict(self) -> dict:
        return {
                'userIri': self.__attr.get(UserAttr.USER_IRI),
                'userId': self.__attr[UserAttr.USER_ID],
                'familyName': self.__attr[UserAttr.FAMILY_NAME],
                'givenName': self.__attr[UserAttr.GIVEN_NAME],
                'isActive': self.__attr[UserAttr.ACTIVE],
                'hasPermissions': self.__attr[UserAttr.HAS_PERMISSIONS],
                'inProject': self.__attr[UserAttr.IN_PROJECT]
        }

    #
    # this private method handles the setting of a field. Whenever a field is being
    # set or modified, this method is called. It also puts the original value and the
    # action into the changeset.
    #
    def __change_setter(self, field: UserAttr, value: UserAttrTypes) -> None:
        if self.__attr[field] == value:
            return
        if self.__attr[field] is None:
            if self.__changeset.get(field) is None:
                self.__changeset[field] = UserAttrChange(None, Action.CREATE)
        else:
            if value is None:
                if self.__changeset.get(field) is None:
                    self.__changeset[field] = UserAttrChange(self.__attr[field], Action.DELETE)
            else:
                if self.__changeset.get(field) is None:
                    self.__changeset[field] = UserAttrChange(self.__attr[field], Action.REPLACE)
        if value is None:
            del self.__attr[field]
            if field == UserAttr.IN_PROJECT:
                self.__attr[field] = InProjectClass(on_change=self.__inProject_cb)
            elif field == UserAttr.HAS_PERMISSIONS:
                self.__attr[field] = ObservableSet(on_change=self.__hasPermission_cb)
        else:
            if field == UserAttr.IN_PROJECT:
                self.__attr[field] = InProjectClass(value, on_change=self.__inProject_cb)
            elif field == UserAttr.HAS_PERMISSIONS:
                self.__attr[field] = ObservableSet(value, on_change=self.__hasPermission_cb)
            else:
                self.__attr[field] = self.__datatypes[field](value)

    #
    # Callbacks for the `ObservableSet`class. This is used whenever the `hasPermission`or
    # `inProject`properties are being modified
    #
    def __hasPermission_cb(self, oldset: ObservableSet, data: Any = None) -> None:
        if self.__changeset.get(UserAttr.HAS_PERMISSIONS) is None:
            self.__changeset[UserAttr.HAS_PERMISSIONS] = UserAttrChange(oldset, Action.MODIFY)

    def __inProject_cb(self, key: Iri, old: ObservableSet[AdminPermission] | None = None) -> None:
        if self.__changeset.get(UserAttr.IN_PROJECT) is None:
            old = None
            if self.__attr.get(UserAttr.IN_PROJECT) is not None:
                old = self.__attr[UserAttr.IN_PROJECT].copy()
            self.__changeset[UserAttr.IN_PROJECT] = UserAttrChange(old, Action.MODIFY)

    @property
    def creator(self) -> Iri | None:
        return self._creator

    @property
    def created(self) -> Xsd_dateTime | None:
        return self._created

    @property
    def contributor(self) -> Iri | None:
        return self._contributor

    @property
    def modified(self) -> Xsd_dateTime | None:
        return self._modified

    @modified.setter
    def modified(self, value: Xsd_dateTime) -> None:
        self._modified = value

    def add_project_permission(self, project: Iri | str, permission: AdminPermission | None) -> None:
        """
        Adds a new administraive permission to the user. If the user is not yet member of the project, he
        will automatically become a member.

        :param project: Name of the project to add the permission
        :type project: Xsd_QName
        :param permission: The admin permission to be added
        :type permission: AdminPermission | None
        :return: None
        """
        if self.__attr[UserAttr.IN_PROJECT].get(project) is None:
            if self.__changeset.get(UserAttr.IN_PROJECT) is None:
                self.__changeset[UserAttr.IN_PROJECT] = UserAttrChange(self.__attr[UserAttr.IN_PROJECT], Action.CREATE)
            self.__attr[UserAttr.IN_PROJECT][project] = ObservableSet({permission})
        else:
            if self.__changeset.get(UserAttr.IN_PROJECT) is None:
                self.__changeset[UserAttr.IN_PROJECT] = UserAttrChange(self.__attr[UserAttr.IN_PROJECT], Action.MODIFY)
            self.__attr[UserAttr.IN_PROJECT][project].add(permission)

    def remove_project_permission(self, project: Iri | str, permission: AdminPermission | None) -> None:
        """
        Remove the given Permission from the user (for the given project)

        :param project: Name of the project
        :type project: Xsd_QName
        :param permission: The permission to be removed
        :type permission: AdminPermission | None
        :return:
        """
        if not isinstance(project, Iri):
            project = Iri(project)
        if self.__attr[UserAttr.IN_PROJECT].get(project) is None:
            raise OldapErrorValue(f"Project '{project}' does not exist")
        if self.__changeset.get(UserAttr.IN_PROJECT) is None:
            self.__changeset[UserAttr.IN_PROJECT] = UserAttrChange(self.__attr[UserAttr.IN_PROJECT], Action.MODIFY)
        self.__attr[UserAttr.IN_PROJECT][project].remove(permission)

    @property
    def changeset(self) -> Dict[UserAttr, UserAttrChange]:
        """
        Return the changeset, that is dict with information about all properties that have benn changed.
        :return: A dictionary of all changes
        """
        return self.__changeset

    def clear_changeset(self):
        """
        Clear the changeset.
        :return:
        """
        self.__changeset = {}

    @staticmethod
    def sparql_query(context: Context, userId: Xsd_NCName) -> str:
        """
        Return the SPARQL query that retrieves the given user from the triple store
        :param context: A `Context` instance
        :type context: Context
        :param userId: The user if
        :type userId: Xsd_NCName
        :return: SPARQL query
        """
        sparql = context.sparql_context
        sparql += f"""
        SELECT ?user ?prop ?val ?proj ?rval
        FROM oldap:admin
        WHERE {{
            {{
                ?user a oldap:User .
                ?user oldap:userId {userId.toRdf} .
                ?user ?prop ?val .
            }} UNION {{
                ?user a oldap:User .
                ?user oldap:userId {userId.toRdf} .
                <<?user oldap:inProject ?proj>> oldap:hasAdminPermission ?rval
            }}
        }}
        """
        return sparql

    def _create_from_queryresult(self, queryresult: QueryProcessor) -> None:
        """
        Create a user from a queryresult created by the method
        :param queryresult:
        :type queryresult: QueryProcessor
        :return: None
        :raises OldapErrorNotFound: Given user not found!
        """
        in_project: Dict[str, Set[AdminPermission]] | None = None
        if len(queryresult) == 0:
            raise OldapErrorNotFound("Given user not found!")
        for r in queryresult:
            match str(r.get('prop')):
                case 'dcterms:creator':
                    self._creator = r['val']
                    self.__attr[UserAttr.USER_IRI] = r['user']
                case 'dcterms:created':
                    self._created = r['val']
                case 'dcterms:contributor':
                    self._contributor = r['val']
                case 'dcterms:modified':
                    self._modified = r['val']
                case 'oldap:userId':
                    self.__attr[UserAttr.USER_ID] = r['val']
                case 'foaf:familyName':
                    self.__attr[UserAttr.FAMILY_NAME] = r['val']
                case 'foaf:givenName':
                    self.__attr[UserAttr.GIVEN_NAME] = r['val']
                case 'oldap:credentials':
                    self.__attr[UserAttr.CREDENTIALS] = r['val']
                case 'oldap:isActive':
                    self.__attr[UserAttr.ACTIVE] = r['val']
                case 'oldap:inProject':
                    in_project = {r['val']: set()}
                    # self.__fields[UserFields.IN_PROJECT] = {str(r['val']): ObservableSet()}
                case 'oldap:hasPermissions':
                    self.__attr[UserAttr.HAS_PERMISSIONS].add(r['val'])
                case _:
                    if r.get('proj') and r.get('rval'):
                        if in_project.get(r['proj']) is None:
                            in_project[r['proj']] = set()
                        in_project[r['proj']].add(AdminPermission(str(r['rval'])))
                        # if self.__fields[UserFields.IN_PROJECT].get(str(r['proj'])) is None:
                        #    self.__fields[UserFields.IN_PROJECT][str(r['proj'])] = ObservableSet()
                        # self.__fields[UserFields.IN_PROJECT][str(r['proj'])].add(AdminPermission(str(r['rval'])))
        if in_project:
            self.__attr[UserAttr.IN_PROJECT] = InProjectClass(in_project, on_change=self.__inProject_cb)
        if not isinstance(self._modified, Xsd_dateTime):
            raise OldapErrorValue(f"Modified field is {type(self._modified)} and not datetime!!!!")
        self.clear_changeset()

    def _sparql_update(self, indent: int = 0, indent_inc: int = 4) -> Tuple[str | None, int, str]:
        """
        return the sparql that performs a sparql query to update all the changes
        :param indent: SPARQL formatting indentation step
        :param indent_inc: SPARQL formatting indentation step size
        :return: A Tuple with three elements:
                 1. SPARQL query to test if the given permission set exists
                 2. The expected length of the result of ptest
                 3. The SPARQL to update the instance
        """
        ptest = None
        ptest_len = 0
        blank = ''
        sparql_list = []
        for field, change in self.__changeset.items():
            if field == UserAttr.HAS_PERMISSIONS or field == UserAttr.IN_PROJECT:
                continue
            sparql = f'{blank:{indent * indent_inc}}# User field "{field.value}" with action "{change.action.value}"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH oldap:admin\n'
            if change.action != Action.CREATE:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?user {field.value} {change.old_value.toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if change.action != Action.DELETE:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}?user {field.value} {self.__attr[field].toRdf} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.userIri.toRdf} as ?user)\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?user {field.value} {change.old_value.toRdf} .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            sparql_list.append(sparql)
        if UserAttr.HAS_PERMISSIONS in self.__changeset:
            new_set = self.__attr[UserAttr.HAS_PERMISSIONS]
            old_set = self.__changeset[UserAttr.HAS_PERMISSIONS].old_value
            added = new_set - old_set
            removed = old_set - new_set
            sparql = f'{blank:{indent * indent_inc}}# User field "hasPermission"\n'
            sparql += f'{blank:{indent * indent_inc}}WITH oldap:admin\n'
            if removed:
                sparql += f'{blank:{indent * indent_inc}}DELETE {{\n'
                for perm in removed:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?user oldap:hasPermissions {perm} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            if added:
                sparql += f'{blank:{indent * indent_inc}}INSERT {{\n'
                for perm in added:
                    sparql += f'{blank:{(indent + 1) * indent_inc}}?user oldap:hasPermissions {perm} .\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
            sparql += f'{blank:{indent * indent_inc}}WHERE {{\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}BIND({self.userIri.toRdf} as ?user)\n'
            sparql += f'{blank:{(indent + 1) * indent_inc}}?user a oldap:User .\n'
            sparql += f'{blank:{indent * indent_inc}}}}'
            if removed or added:
                sparql_list.append(sparql)

            #
            # check if existing :PermissionSet's have been given!
            #
            if added:
                ptest = f"""
                SELECT ?permissionset
                FROM oldap:admin
                WHERE {{
                    ?permissionset a oldap:PermissionSet .
                    FILTER(?permissionset IN ({added.toRdf}))
                }}
                """
                ptest_len = len(added) if added else 0

        if UserAttr.IN_PROJECT in self.__changeset:
            # first get all keys that must be added, that is that are in NEW but not in OLD:
            addedprojs = self.__attr[UserAttr.IN_PROJECT].keys() - self.__changeset[UserAttr.IN_PROJECT].old_value.keys()
            deletedprojs = self.__changeset[UserAttr.IN_PROJECT].old_value.keys() - self.__attr[UserAttr.IN_PROJECT].keys()
            changedprojs = self.__attr[UserAttr.IN_PROJECT].keys() & self.__changeset[UserAttr.IN_PROJECT].old_value.keys()

            # add projects
            if addedprojs:
                sparql = f"{blank:{indent * indent_inc}}INSERT DATA {{\n"
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                for proj in addedprojs:
                    sparql += f'{blank:{(indent + 2) * indent_inc}}{self.userIri.toRdf} oldap:inProject {proj.toRdf} .\n'
                    for perm in self.__attr[UserAttr.IN_PROJECT][proj]:
                        sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                sparql_list.append(sparql)

            # delete projects
            if deletedprojs:
                sparql = f"{blank:{indent * indent_inc}}DELETE DATA {{\n"
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                for proj in deletedprojs:
                    sparql += f'{blank:{(indent + 2) * indent_inc}}{self.userIri.toRdf} oldap:inProject {proj.toRdf} .\n'
                    for perm in self.__changeset[UserAttr.IN_PROJECT].old_value[proj]:
                        sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                sparql_list.append(sparql)

            if changedprojs:
                doit = False
                sparql = f"{blank:{indent * indent_inc}}INSERT DATA {{\n"
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                for proj in changedprojs:
                    perms = self.__attr[UserAttr.IN_PROJECT][proj] - self.__changeset[UserAttr.IN_PROJECT].old_value[proj]
                    for perm in perms:
                        sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                        doit = True
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                if doit:
                    sparql_list.append(sparql)

                doit = False
                sparql = f"{blank:{indent * indent_inc}}DELETE DATA {{\n"
                sparql += f'{blank:{(indent + 1) * indent_inc}}GRAPH oldap:admin {{\n'
                for proj in changedprojs:
                    perms = self.__changeset[UserAttr.IN_PROJECT].old_value[proj] - self.__attr[UserAttr.IN_PROJECT][proj]
                    for perm in perms:
                        sparql += f'{blank:{(indent + 2) * indent_inc}}<<{self.userIri.toRdf} oldap:inProject {proj.toRdf}>> oldap:hasAdminPermission {perm.value} .\n'
                        doit = True
                sparql += f'{blank:{(indent + 1) * indent_inc}}}}\n'
                sparql += f'{blank:{indent * indent_inc}}}}\n'
                if doit:
                    sparql_list.append(sparql)
        return ptest, ptest_len, " ;\n".join(sparql_list)


if __name__ == "__main__":
    gaga = UserDataclass()
    user_dataclass = UserDataclass(
        userIri=Iri("https://orcid.org/0000-0002-9991-2055"),
        userId=Xsd_NCName("edison"),
        familyName="Edison",
        givenName="Tholdap A.",
        credentials="Lightbulb&Phonograph",
        inProject={Iri('oldap:HyperHamlet'): {AdminPermission.ADMIN_USERS,
                                                   AdminPermission.ADMIN_RESOURCES,
                                                   AdminPermission.ADMIN_CREATE}},
        hasPermissions={Xsd_QName('oldap:GenericView')})
    print(user_dataclass.userId)
